# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest

from nark.backends.sqlalchemy.objects import AlchemyActivity, AlchemyCategory
from nark.items.activity import Activity


class TestActivityManager:
    """"""

    def test_get_or_create_get(self, alchemy_store, alchemy_activity):
        """
        Make sure that passing an existing activity retrieves the corresponding
        instance.

        Note:
            * The activity will is be looked up by its composite key, so not to
            make any assumptions on the existence of a PK.
        """
        activity = alchemy_activity.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        result = alchemy_store.activities.get_or_create(activity)
        assert result == activity
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_get_or_create_new(self, alchemy_store, activity):
        """
        Make sure that passing a new activity create a new persitent instance.

        Note:
            * The activity will is be looked up by its composite key, so not to
            make any assumptions on the existence of a PK.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        result = alchemy_store.activities.get_or_create(activity)
        assert result.equal_fields(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_save_new(self, activity, alchemy_store):
        """Make sure that saving a new activity add a new persistent instance."""
        # [TODO]
        # This should not be needed as ``save`` is a basestore method.
        # Its just a case of 'better save than sorry.
        assert activity.pk is None
        count_before = alchemy_store.session.query(AlchemyActivity).count()
        result = alchemy_store.activities._add(activity)
        count_after = alchemy_store.session.query(AlchemyActivity).count()
        assert count_before < count_after
        assert result.equal_fields(activity)

    def test_save_existing(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_category_factory,
    ):
        """
        Make sure that saving an existing activity add no new persistent
        instance.
        """
        # [TODO]
        # This should not be needed as ``save`` is a basestore method.
        activity = alchemy_activity.as_hamster(alchemy_store)
        activity.category = alchemy_category_factory()
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2
        result = alchemy_store.activities.save(activity)
        assert result == activity
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2

    def test_activity_without_category(self, alchemy_store, activity):
        """Add a new activity without an category."""
        activity.category = None
        result = alchemy_store.activities._add(activity)
        assert result.equal_fields(activity)

    def test_add_new_with_new_category(self, alchemy_store, activity, category):
        """
        Test that adding a new alchemy_activity with new alchemy_category
        creates both.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        activity.category = category
        result = alchemy_store.activities._add(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_add_new_with_existing_category(
        self,
        alchemy_store,
        activity,
        alchemy_category,
    ):
        """
        Test that adding a new activity with existing category does not
        create a new one.
        """
        activity.category = alchemy_category.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        result = alchemy_store.activities._add(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_add_new_with_existing_name_and_alchemy_category(
        self,
        alchemy_store,
        activity,
        alchemy_activity,
    ):
        """
        Test that adding a new alchemy_activity_with_existing_composite_key
        throws error.
        """
        activity.name = alchemy_activity.name
        activity.category = alchemy_activity.category.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        with pytest.raises(ValueError):
            alchemy_store.activities._add(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_add_with_pk(self, alchemy_store, activity):
        """Make sure that adding an alchemy_activity with a PK raises error."""
        activity.pk = 234
        with pytest.raises(ValueError):
            alchemy_store.activities._add(activity)

    # *** ._update() method tests.

    def test_update_without_pk(self, alchemy_store, activity):
        """Make sure that calling update without a PK raises exception."""
        with pytest.raises(ValueError) as excinfo:
            alchemy_store.activities._update(activity)
        assert str(excinfo.value).startswith("The Activity passed")

    def test_update_with_existing_name_and_new_category_name(
        self,
        alchemy_store,
        activity,
        alchemy_activity,
        alchemy_category_factory,
    ):
        """
        Make sure that calling update on existing composite key raises exception.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

        assert activity.pk is None
        activity.pk = alchemy_activity.pk
        activity.name = alchemy_activity.name

        category = alchemy_category_factory()
        assert alchemy_activity.category != category
        assert activity.category.pk is None
        # And now for something completely different.
        activity.category.name = category.name

        result = alchemy_store.activities._update(activity)
        # The result is almost the same as what we gave it -- except now
        # the category PK will have been assigned, check it:
        activity.category.pk = result.category.pk
        assert result == activity

    def test_update_fails_on_existing_activity_category_conflict(
        self,
        alchemy_store,
        activity,
        alchemy_activity,
        alchemy_category,
    ):
        """
        Make sure that calling update on existing composite key raises exception.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        # (lb): Interesting. Or maybe not? Did not expect fixture item object to be
        # same as the one in the fixture store, especially because both the activity
        # and the category exist in store whether or not also specified as a fixture.
        stored_activity = alchemy_store.session.query(AlchemyActivity).one()
        assert id(stored_activity) == id(alchemy_activity)
        assert id(stored_activity.category) == id(alchemy_category)

        assert activity.pk is None
        activity.pk = alchemy_activity.pk
        activity.name = alchemy_activity.name
        activity.category.name = alchemy_category.name

        with pytest.raises(ValueError) as excinfo:
            alchemy_store.activities._update(activity)
        assert str(excinfo.value).startswith("The database already contains that")

    def test_update_with_existing_category_name_conflict_again(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_category_factory,
    ):
        activity = alchemy_activity.as_hamster(alchemy_store)
        category = alchemy_category_factory().as_hamster(alchemy_store)
        activity.category = category
        alchemy_store.activities._update(activity)
        with pytest.raises(ValueError) as excinfo:
            # 2nd time's a no-go.
            alchemy_store.activities._update(activity)
        assert excinfo.value.args[0].startswith("The database already contains that")

    def test_update_with_invalid_pk(self, alchemy_store, activity, alchemy_activity):
        """Make sure that calling update without a PK raises exception."""
        activity.pk = alchemy_activity.pk + 1
        with pytest.raises(KeyError) as excinfo:
            alchemy_store.activities._update(activity)
        assert excinfo.value.args[0].startswith("No Activity with PK")

    def test_update_with_existing_category(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_category_factory,
    ):
        """
        Test that updateting an activity with existing category does not
        create a new one.
        """
        activity = alchemy_activity.as_hamster(alchemy_store)
        category = alchemy_category_factory().as_hamster(alchemy_store)
        assert alchemy_activity.category != category
        activity.category = category
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2
        result = alchemy_store.activities._update(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_update_name(
        self,
        alchemy_store,
        alchemy_activity,
        name_string_valid_parametrized,
    ):
        """Test updateing an activities name with a valid new string."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        activity.name = name_string_valid_parametrized
        result = alchemy_store.activities._update(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    # ***

    def test_remove_existing(self, alchemy_store, alchemy_activity):
        """Make sure removing an existsing alchemy_activity works as intended."""
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        activity = alchemy_activity.as_hamster(alchemy_store)
        alchemy_store.activities.remove(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 0

    def test_remove_no_pk(self, alchemy_store, activity):
        """
        Make sure that trying to remove an alchemy_activity without a PK
        raises error.
        """
        with pytest.raises(ValueError):
            alchemy_store.activities.remove(activity)

    def test_remove_invalid_pk(self, alchemy_store, alchemy_activity):
        """Test that removing of a non-existent key raises error."""
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        activity = alchemy_activity.as_hamster(alchemy_store)
        activity.pk = activity.pk + 1
        with pytest.raises(KeyError):
            alchemy_store.activities.remove(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1

    def test_get_existing(self, alchemy_store, alchemy_activity):
        """
        Make sure that retrieving an existing alchemy_activity by pk works
        as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity.pk)
        assert result == alchemy_activity.as_hamster(alchemy_store)
        assert result is not alchemy_activity

    def test_get_existing_raw(self, alchemy_store, alchemy_activity):
        """
        Make sure that retrieving an existing alchemy_activity by pk works
        as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity.pk, raw=True)
        assert result == alchemy_activity
        assert result is alchemy_activity

    # FIXME: Add deleted/hidden tests...
    # FIXME: Split this file into one file for each item type.
    def test_get_existing_deleted_false(self, alchemy_store, alchemy_activity):
        """
        Make sure that retrieving an existing alchemy_activity by
        deleted=False works as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity.pk, deleted=False)
        assert result == alchemy_activity.as_hamster(alchemy_store)
        assert result is not alchemy_activity

    def test_get_existing_deleted_true(self, alchemy_store, alchemy_activity_deleted):
        """
        Make sure that retrieving an existing alchemy_activity by
        deleted=True works as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity_deleted.pk, deleted=True)
        assert result == alchemy_activity_deleted.as_hamster(alchemy_store)
        assert result is not alchemy_activity_deleted

    def test_get_non_existing(self, alchemy_store):
        """Make sure quering for a non existent PK raises error."""
        with pytest.raises(KeyError):
            alchemy_store.activities.get(4)

    @pytest.mark.parametrize("raw", (True, False))
    def test_get_by_composite_valid(self, alchemy_store, alchemy_activity, raw):
        """Make sure that querying for a valid name/alchemy_category combo succeeds."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        result = alchemy_store.activities.get_by_composite(
            activity.name,
            activity.category,
            raw=raw,
        )
        if raw:
            assert result == alchemy_activity
            assert result is alchemy_activity
        else:
            assert result == alchemy_activity
            assert result is not alchemy_activity

    def test_get_by_composite_invalid_category(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_category_factory,
    ):
        """Make sure that querying with an invalid category raises errror."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        category = alchemy_category_factory().as_hamster(alchemy_store)
        with pytest.raises(KeyError):
            alchemy_store.activities.get_by_composite(activity.name, category)

    def test_get_by_composite_invalid_name(
        self,
        alchemy_store,
        alchemy_activity,
        name_string_valid_parametrized,
    ):
        """Make sure that querying with an invalid alchemy_category raises errror."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        invalid_name = activity.name + "foobar"
        with pytest.raises(KeyError):
            alchemy_store.activities.get_by_composite(invalid_name, activity.category)

    def test_get_all_without_category(self, alchemy_store, alchemy_activity):
        """Make sure method returns all activities."""
        results = alchemy_store.activities.get_all()
        assert len(results) == 1

    def test_get_all_with_category_none(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_activity_factory,
    ):
        """Make sure only activity without a category is returned."""
        activity_without_category = alchemy_activity_factory(category=None)
        results = alchemy_store.activities.get_all(match_categories=[None])
        assert len(results) == 1
        assert results[0] == activity_without_category

    def test_get_all_with_category_named(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_activity_factory,
    ):
        """Make sure only activity matching the given alchemy_category is returned."""
        # Add Activity without Category to data store.
        # F841 local variable 'foo' is assigned to but never used
        _activity_sans_category = alchemy_activity_factory(category=None)  # noqa: F841
        results = alchemy_store.activities.get_all(
            match_categories=[alchemy_activity.category],
        )
        assert len(results) == 1
        assert results[0] == alchemy_activity

    def test_get_all_with_category_both_none_and_named(
        self,
        alchemy_store,
        alchemy_activity,
        alchemy_activity_factory,
    ):
        """Make sure both activities with category name and with none are returned."""
        # Add Activity without Category to data store.
        _activity_sans_category = alchemy_activity_factory(category=None)  # noqa: F841
        results = alchemy_store.activities.get_all(
            match_categories=[None, alchemy_activity.category],
        )
        assert len(results) == 2

    def test_get_all_with_search_term(self, alchemy_store, alchemy_activity):
        """
        Make sure that activities matching the given term ass name are returned.
        """
        # (lb): This test previously hydrated the AlchemyActivity into a proper
        # Activity, but it passes fine without. So not sure if I'm missing
        # something, but we can skip this step:
        #   activity = alchemy_activity.as_hamster(alchemy_store)
        # (and could probably do the same elsewhere in this file, but not
        # really worth the time -- just something I wanted to note).
        results = alchemy_store.activities.get_all(
            match_categories=[alchemy_activity.category],
            search_terms=[alchemy_activity.name],
        )
        assert len(results) == 1

    def test_get_all_with_category_miss(self, alchemy_store, alchemy_activity):
        """
        Make sure that activities matching the given alchemy_category are returned.
        """
        results = alchemy_store.activities.get_all(match_categories=["miss"])
        assert len(results) == 0
        assert alchemy_activity == alchemy_store.activities.get_all()[0]

    def test_get_deleted_item(self, alchemy_store, alchemy_activity):
        """Make sure method retrieves deleted object."""
        alchemy_activity.deleted = True
        result = alchemy_store.activities.get(alchemy_activity.pk, deleted=True)
        assert result == alchemy_activity

    def test_get_all_match_activities(
        self,
        alchemy_store,
        set_of_alchemy_facts_active,
        alchemy_activity_factory,
    ):
        """Test get_all argument: QueryTerms.match_activities."""
        # Cover as many branches as possible within query_filter_by_activity
        # and query_filter_by_activity_name.
        activity_0 = set_of_alchemy_facts_active[0].activity
        activity_2 = set_of_alchemy_facts_active[2].activity.name
        # Rather than use alchemy_activity or alchemy_activity_factory,
        # which will insist that PK by non-None (though possibly 0 is okay,
        # and we could test the same path...), use a normal Activity.
        # This will assign a new PK:
        any_activity = alchemy_activity_factory(pk=None)
        activity_with_pk_None = Activity(name=any_activity.name)
        results = alchemy_store.activities.get_all(
            match_activities=[
                activity_0,
                activity_2,
                activity_with_pk_None,
                None,
            ],
        )
        assert len(results) == 3

    # ***

    @pytest.mark.parametrize(
        ("sort_cols"),
        (
            (["start"]),
            ([None]),
            (["usage"]),
            (["time"]),
            (["activity"]),
            (["category"]),
        ),
    )
    def test_get_all_sort_cols(self, alchemy_store, sort_cols):
        alchemy_store.activities.get_all(sort_cols=sort_cols)

    def test_get_all_sort_cols_tag_fails(self, alchemy_store, mocker):
        # The tag table is not joined for Activity query.
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.activities.get_all(sort_cols=["tag"])
        assert alchemy_store.logger.warning.called

    def test_get_all_sort_cols_unknown(self, alchemy_store, mocker):
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.activities.get_all(sort_cols=["foo"])
        assert alchemy_store.logger.warning.called



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from datetime import datetime
from gettext import gettext as _

from sqlalchemy import asc, desc, func
from sqlalchemy.sql.expression import and_, or_

from ..objects import AlchemyFact
from . import query_apply_true_or_not, query_prepare_datetime
from .gather_fact import GatherFactManager

__all__ = ("FactManager",)


class FactManager(GatherFactManager):
    """ """

    def __init__(self, *args, **kwargs):
        super(FactManager, self).__init__(*args, **kwargs)

    # ***

    def _add(self, fact, raw=False, skip_commit=False, ignore_pks=[]):
        """
        Add a new fact to the database.

        Args:
            fact (nark.Fact): Fact to be added.
            raw (bool): If ``True`` return ``AlchemyFact`` instead.

        Returns:
            nark.Fact: Fact as stored in the database

        Raises:
            ValueError: If the passed fact has a PK assigned.
                New facts should not have one.

            ValueError: If the time window is already occupied.
        """
        self.adding_item_must_not_have_pk(fact)

        self.must_validate_datetimes(fact, ignore_pks=ignore_pks)

        alchemy_fact = AlchemyFact(
            pk=None,
            activity=None,
            # FIXME/2018-08-23 00:38: Is this still valid?
            # FIXME: mircoseconds are being stored...
            #        I modified fact_time.must_be_datetime_or_relative to strip
            #        milliseconds. but they're still being saved (just as six 0's).
            start=fact.start,
            end=fact.end,
            description=fact.description,
            deleted=bool(fact.deleted),
            split_from=fact.split_from,
        )
        get_or_create = self.store.activities.get_or_create
        alchemy_fact.activity = get_or_create(fact.activity, raw=True, skip_commit=True)
        tags = [
            self.store.tags.get_or_create(tag, raw=True, skip_commit=True)
            for tag in fact.tags
        ]
        alchemy_fact.tags = tags

        result = self.add_and_commit(
            alchemy_fact,
            raw=raw,
            skip_commit=skip_commit,
        )

        return result

    # ***

    def _update(self, fact, raw=False, ignore_pks=[]):
        """
        Update and existing fact with new values.

        Args:
            fact (nark.fact): Fact instance holding updated values.

            raw (bool): If ``True`` return ``AlchemyFact`` instead.
              ANSWER: (lb): "instead" of what? raw is not used by Fact...

        Returns:
            nark.fact: Updated Fact

        Raises:
            KeyError: if a Fact with the relevant PK could not be found.
            ValueError: If the the passed activity does not have a PK assigned.
            ValueError: If the time window is already occupied.
        """
        self.store.logger.debug("Received: {!r} / raw: {}".format(fact, raw))

        if not fact.pk:
            message = _(
                "The Fact passed ('{!r}') does not have a PK."
                " We do not know which entry to modify."
            ).format(fact)
            self.store.logger.error(message)
            raise ValueError(message)

        self.must_validate_datetimes(fact, ignore_pks=ignore_pks)

        alchemy_fact = self.store.session.query(AlchemyFact).get(fact.pk)
        if not alchemy_fact:
            message = _("No Fact with PK ‘{}’ was found.").format(fact.pk)
            self.store.logger.error(message)
            raise KeyError(message)

        if alchemy_fact.deleted:
            message = _("Cannot edit deleted Fact: ‘{!r}’.".format(fact))
            self.store.logger.error(message)
            raise ValueError(message)

        if (
            (fact.deleted and (fact.end == alchemy_fact.end))
            or (fact.end and not alchemy_fact.end)
        ) and fact.equal_sans_end(alchemy_fact):
            # Don't bother with split_from entry.
            # MAYBE: (lb): Go full wiki and store edit times? Ug...
            new_fact = alchemy_fact
            alchemy_fact.deleted = fact.deleted
            alchemy_fact.end = fact.end
        else:
            assert alchemy_fact.pk == fact.pk
            was_split_from = fact.split_from
            fact.split_from = alchemy_fact
            # Clear the ID so that a new ID is assigned.
            fact.pk = None
            new_fact = self._add(
                fact, raw=True, skip_commit=True, ignore_pks=ignore_pks
            )
            # NOTE: _add() calls:
            #       self.store.session.commit()
            # The fact being split from is deleted/historic.
            alchemy_fact.deleted = True
            assert new_fact.pk > alchemy_fact.pk
            # Restore the ID to not confuse the caller!
            # The caller will still have a handle on Fact. Rather than
            # change its pk to new_fact's, have it reflect its new
            # split_from status.
            fact.pk = alchemy_fact.pk
            fact.split_from = was_split_from
            # The `alchemy_fact` is what gets saved, but the `fact`
            # is what the caller passed us, so update it, too.
            fact.deleted = True

        self.store.session.commit()

        self.store.logger.debug("Updated: {!r}".format(fact))

        if not raw:
            new_fact = new_fact.as_hamster(self.store)

        return new_fact

    # ***

    def must_validate_datetimes(self, fact, ignore_pks=[]):
        if not isinstance(fact.start, datetime):
            raise TypeError(_("Missing start time for ‘{!r}’.").format(fact))

        # Check for valid time range.
        invalid_range = False
        if fact.end is not None:
            if fact.start > fact.end:
                invalid_range = True
            else:
                # EXPERIMENTAL: Sneaky, "hidden", vacant, timeless Facts.
                allow_momentaneous = self.store.config["time.allow_momentaneous"]
                if not allow_momentaneous and fact.start >= fact.end:
                    invalid_range = True

        if invalid_range:
            message = _("Invalid time range for “{!r}”.").format(fact)
            if fact.start == fact.end:  # pragma: no cover
                assert False  # (lb): Preserved in case we revert == policy.
                message += _(" The start time cannot be the same as the end time.")
            else:
                message += _(" The start time cannot be after the end time.")
            self.store.logger.error(message)
            raise ValueError(message)

        if not self._timeframe_available_for_fact(fact, ignore_pks):
            msg = _(
                "One or more Facts already exist "
                "between the indicated start and end times. "
            )
            self.store.logger.error(msg)
            raise ValueError(msg)

    # ***

    def _timeframe_available_for_fact(self, fact, ignore_pks=[]):
        """
        Determine if a timeframe given by the passed fact is already occupied.

        This method takes also such facts into account that start before and end
        after the fact in question. In that regard it exceeds what ``gather``
        would return.

        Args:
            fact (Fact): The fact to check. Please note that the fact is expected to
                have a ``start`` and ``end``.

        Returns:
            bool: ``True`` if the timeframe is available, ``False`` if not.

        Note:
            If the given fact is the only fact instance within the given timeframe
            the timeframe is considered available (for this fact)!
        """
        # Use func.datetime and query_prepare_datetime to normalize time comparisons,
        # so that equivalent times that are expressed differently are evaluated
        # as equal, e.g., "2018-01-01 10:00" should match "2018-01-01 10:00:00".
        # FIXME: func.datetime is SQLite-specific: need to abstract for other DBMSes.

        start = query_prepare_datetime(fact.start)
        query = self.store.session.query(AlchemyFact)

        # FIXME: Only use func.datetime on SQLite store.
        #
        #   (lb): SQLite stores datetimes as strings, so what's in the store
        #   might vary depending on, say, changes to this code. As such, some
        #   start and end times might include seconds, and some times might not.
        #   Here we use func.datetime and query_prepare_datetime to normalize the
        #   comparison. But this is SQLite-specific, so we should abstract
        #   the operation for other DBMSes (and probably do nothing, since most
        #   other databases have an actual datetime data type).
        condition = and_(func.datetime(AlchemyFact.end) > start)
        if fact.end is not None:
            end = query_prepare_datetime(fact.end)
            condition = and_(condition, func.datetime(AlchemyFact.start) < end)
        else:
            # The fact is ongoing, so match the ongoing (active) Fact in the store.
            # E711: `is None` breaks Alchemy, so use `== None`.
            condition = or_(AlchemyFact.end == None, condition)  # noqa: E711

        if fact.pk:
            condition = and_(condition, AlchemyFact.pk != fact.pk)

        if fact.split_from:
            condition = and_(condition, AlchemyFact.pk != fact.split_from.pk)

        if ignore_pks:
            condition = and_(condition, AlchemyFact.pk.notin_(ignore_pks))

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        n_found = query.count()

        return not bool(n_found)

    # ***

    def remove(self, fact, purge=False):
        """
        Remove a fact from our internal backend.

        Args:
            fact (nark.Fact): Fact to be removed

        Returns:
            bool: Success status

        Raises:
            ValueError: If fact passed does not have an pk.

            KeyError: If no fact with passed PK was found.
        """
        self.store.logger.debug("Received: {!r}".format(fact))

        if not fact.pk:
            message = _(
                "Fact ‘{!r}’ has no PK. Are you trying to remove a new Fact?"
            ).format(fact)
            self.store.logger.error(message)
            raise ValueError(message)

        alchemy_fact = self.store.session.query(AlchemyFact).get(fact.pk)
        if not alchemy_fact:
            message = _("No Fact with PK ‘{}’ was found.").format(fact.pk)
            self.store.logger.error(message)
            raise KeyError(message)
        if alchemy_fact.deleted:
            message = _("The Fact is already marked deleted.")
            self.store.logger.error(message)
            # FIXME/2018-06-08: (lb): I think we need custom Exceptions...
            raise Exception(message)
        alchemy_fact.deleted = True
        if purge:
            self.store.session.delete(alchemy_fact)
        self.store.session.commit()
        self.store.logger.debug("Deleted: {!r}".format(fact))

    # ***

    def get(self, pk, deleted=None, raw=False):
        """
        Retrieve a fact based on its PK.

        Args:
            pk (int): PK of the fact to be retrieved.

            deleted (boolean, optional):
                False to restrict to non-deleted Facts;
                True to find only those marked deleted;
                None to find all.

            raw (bool): Return the AlchemyActivity instead.

        Returns:
            nark.Fact: Fact matching given PK

        Raises:
            KeyError: If no Fact of given key was found.
        """
        self.store.logger.debug("Received PK: ‘{}’ / raw: {}.".format(pk, raw))

        if deleted is None:
            query = self.store.session.query(AlchemyFact)
            result = query.get(pk)
        else:
            query = self.store.session.query(AlchemyFact)
            query = query.filter(AlchemyFact.pk == pk)
            query = query_apply_true_or_not(query, AlchemyFact.deleted, deleted)
            results = query.all()
            assert len(results) <= 1
            result = results[0] if results else None

        if not result:
            message = _("No Fact with PK ‘{}’ was found.").format(pk)
            self.store.logger.error(message)
            raise KeyError(message)
        if not raw:
            # Explain: Why is as_hamster optionable, when act/cat/tag do it always?
            result = result.as_hamster(self.store)
        self.store.logger.debug("Returning: {!r}".format(result))
        return result

    # ***

    def get_all(self, query_terms=None, lazy_tags=False, **kwargs):
        query_terms, kwargs = self._gather_prepare_query_terms(query_terms, **kwargs)
        if query_terms.sort_cols is None:
            query_terms.sort_cols = ("start",)
        return super(FactManager, self).get_all(
            query_terms, lazy_tags=lazy_tags, **kwargs
        )

    # ***

    def get_all_by_usage(self, query_terms=None, **kwargs):
        """Raises if called, because base class defines method for non-Fact item."""
        raise NotImplementedError

    # ***

    def query_exclude_fact(self, query, fact=None):
        if fact and not fact.unstored:
            query = query.filter(AlchemyFact.pk != fact.pk)
        return query

    # ***

    def starting_at(self, fact):
        """
        Return the fact starting at the moment in time indicated by fact.start.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            ValueError: If more than one Fact found at given time.
        """
        query = self.store.session.query(AlchemyFact)

        if fact.start is None:
            raise ValueError("No `start` for starting_at(fact).")

        start_at = query_prepare_datetime(fact.start)
        condition = and_(func.datetime(AlchemyFact.start) == start_at)

        # Excluded 'deleted' Facts.
        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712
        query = query.filter(condition)
        # Exclude fact.pk from results.
        query = self.query_exclude_fact(query, fact)
        # Order by (start time, end time, fact ID), ascending.
        query = self.query_order_by_start(query, asc)

        self.store.logger.debug("fact: {} / query: {}".format(fact, str(query)))

        n_facts = query.count()
        if n_facts > 1:
            message = (
                'More than one fact found starting at "{}": {} facts found'.format(
                    fact.start, n_facts
                )
            )
            raise ValueError(message)

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def ending_at(self, fact):
        """
        Return the fact ending at the moment in time indicated by fact.end.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            ValueError: If more than one Fact found at given time.
        """
        query = self.store.session.query(AlchemyFact)

        if fact.end is None:
            raise ValueError("No `end` for ending_at(fact).")

        end_at = query_prepare_datetime(fact.end)
        condition = and_(func.datetime(AlchemyFact.end) == end_at)

        # Excluded 'deleted' Facts.
        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712
        query = query.filter(condition)
        # Exclude fact.pk from results.
        query = self.query_exclude_fact(query, fact)
        # Order by (start time, end time, fact ID), descending.
        query = self.query_order_by_start(query, desc)

        self.store.logger.debug("fact: {} / query: {}".format(fact, str(query)))

        n_facts = query.count()
        if n_facts > 1:
            message = 'More than one fact found ending at "{}": {} facts found'.format(
                fact.end,
                n_facts,
            )
            raise ValueError(message)

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def antecedent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately preceding the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The antecedent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        query = self.store.session.query(AlchemyFact)

        if fact is not None:
            if fact.end and isinstance(fact.end, datetime):
                # A Closed Fact.
                ref_time = fact.end
            elif fact.start and isinstance(fact.start, datetime):
                # The Active Fact.
                ref_time = fact.start
        if not isinstance(ref_time, datetime):
            # This branch reachable from factoid_fixture tests. See:
            #   raw_fact: 'Monday-13:00: foo@bar', time_hint: 'verify_both'
            #   raw_fact: '2015-12-12 13:00 foo@bar', time_hint: 'verify_both'
            raise ValueError(_("No reference time for antecedent(fact)."))

        ref_time = query_prepare_datetime(ref_time)

        before_active_fact_start = and_(
            AlchemyFact.end == None,  # noqa: E711
            # Except rather than <=, use less than, otherwise
            # penultimate_fact.antecedent might find the ultimate
            # fact, if that final fact is ongoing.
            #   E.g., considering
            #     fact  1: time-a to time-b
            #     ...
            #     fact -2: time-x to time-y
            #     fact -1: time-y to <now>
            #   antecedent of fact -2 should check time-y < time-y and
            #   not <= otherwise antecedent of fact -2 would be fact -1.
            #   (The subsequent function will see it, though, as it
            #   looks for AlchemyFact.start >= ref_time.)
            func.datetime(AlchemyFact.start) < ref_time,
        )

        # (lb): This intricate query is meant to handle momentaneous Facts. If
        # there were no such Facts, we could check AlchemyFact.end <= ref_time
        # and exclude AlchemyFact.pk == fact.pk, and we'd be fine. But we want
        # to handle momentaneous Facts, so we have 3 ref_time comparisons, not 1.
        # - Note that I originally had an if-else block here: if fact is None,
        #   the code would simply check AlchemyFact.end <= ref_time. The use
        #   case here would be passing a datetime to this method, antecedent,
        #   and not a Fact. For example, if you had a Fact from 11a to 12p,
        #   and a Fact from 12p to 12p, passing a datetime set to 12p would
        #   return whichever of those two Facts has the greater PK (because
        #   of the query_order_by_start later in this method). If that Fact
        #   was the momentaneous Fact, you wouldn't be able to call antecedent
        #   again to get the earlier Fact given the datetime alone -- each time
        #   you pass 12p to this method, you'd get the same Fact. But that's not
        #   how this method would be used -- to handle momentaneous Facts, we
        #   need the Fact object, because we need to compare PKs. So rather
        #   than return a momentaneous Fact when a datetime is passed, skip
        #   them. That is, in the example just given, if given a datetime at
        #   12p, always return the Fact from 11a to 12p. We do this by not
        #   just searching AlchemyFact.end <= ref_time, but instead by using
        #   two search criteria, and checking both AlchemyFact.end < ref_time,
        #   or AlchemyFact.end == ref_time and AlchemyFact.start < ref_time.

        or_criteria = []
        # Get complicated, so we handle momentaneous Facts appropriately.
        # E.g., suppose one Fact's time range is 11a to noon, and another
        # Fact is momentaneous from 12:00:00 to 12:00:00. If user is viewing
        # a third Fact that spans from 12:00:00 to 13:00:00, antecedent
        # should return the momentaneous Fact. If antecedent is called
        # again on the momentaneous Fact, return the one that starts at 11a.
        # Start by including any Fact that ends *before*, but not at, ref_time.
        or_criteria.append(func.datetime(AlchemyFact.end) < ref_time)
        # Next, include any Fact that ends at ref_time but is not momentaneous.
        # Given the previous example of three Facts, given the momentaneous
        # Fact at 12:00:00, this will find the earlier Fact from 11a to 12p.
        or_criteria.append(
            and_(
                func.datetime(AlchemyFact.end) == ref_time,
                func.datetime(AlchemyFact.start) < ref_time,
            )
        )
        # Finally, include any momentaneous Fact that occupies the moment at
        # ref_time, but take into consideration the PK so that calling this
        # method, antecedent, with each momentaneous Fact will return them in
        # a predictable order, and will eventually walk out of the moment.
        if fact is not None and fact.pk is not None:
            # From example, assume there are 2 momentaneous Facts at 12:00:00,
            # this would ensure that, after finding the first one, passing the
            # first one to this method finds the second on, and then passing the
            # second one does not return the first one again. (Note that later
            # we call query_order_by_start to ensure the order is correct.)
            or_criteria.append(
                and_(
                    func.datetime(AlchemyFact.end) == ref_time,
                    func.datetime(AlchemyFact.start) == ref_time,
                    AlchemyFact.pk < fact.pk,
                )
            )
        before_closed_fact_end = and_(
            AlchemyFact.end != None,  # noqa: E711
            or_(*or_criteria),
        )

        condition = or_(
            before_active_fact_start,
            before_closed_fact_end,
        )

        # Excluded 'deleted' Facts.
        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712
        query = query.filter(condition)
        # Exclude fact.pk from results.
        query = self.query_exclude_fact(query, fact)
        # Order by (start time, end time, fact ID), descending.
        query = self.query_order_by_start(query, desc)

        query = query.limit(1)

        self.store.logger.debug(
            "fact: {} / ref_time: {} / query: {}".format(fact, ref_time, str(query))
        )

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def subsequent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately following the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The subsequent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        query = self.store.session.query(AlchemyFact)

        if fact is not None:
            if fact.start and isinstance(fact.start, datetime):
                ref_time = fact.start
            elif fact.end and isinstance(fact.end, datetime):
                # A Fact whose start None or not Datetime, but has an end.
                # - E.g., start is None or '+10m'; end is Datetime.
                # - See factoid_fixture -- a number of tests branch here.
                #     py.test -x tests/facts/test_add_fact.py
                #     py.test -x -s -k test_add_new_fact tests/
                ref_time = fact.end
        if ref_time is None:
            raise ValueError(_("No reference time for subsequent(fact)."))

        ref_time = query_prepare_datetime(ref_time)

        # See comments in antecedent that explain the logic here (albeit
        # the complementary logic, for searching backwards, not forward).
        or_criteria = []
        or_criteria.append(func.datetime(AlchemyFact.start) > ref_time)
        or_criteria.append(
            and_(
                func.datetime(AlchemyFact.start) == ref_time,
                func.datetime(AlchemyFact.end) > ref_time,
            )
        )
        if fact is not None and fact.pk is not None:
            or_criteria.append(
                and_(
                    func.datetime(AlchemyFact.start) == ref_time,
                    func.datetime(AlchemyFact.end) == ref_time,
                    AlchemyFact.pk > fact.pk,
                )
            )
        # Note that, by design, AlchemyFact.start should always be not None,
        # but we'll check anyway, for completeness and comparability to the
        # antecedent method (where AlchemyFact.end is not always not None).
        condition = and_(
            AlchemyFact.start != None,  # noqa: E711
            or_(*or_criteria),
        )

        # Excluded 'deleted' Facts.
        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712
        query = query.filter(condition)
        # Exclude fact.pk from results.
        query = self.query_exclude_fact(query, fact)
        # Order by (start time, end time, fact ID), ascending.
        query = self.query_order_by_start(query, asc)

        query = query.limit(1)

        self.store.logger.debug(
            "fact: {} / ref_time: {} / query: {}".format(fact, ref_time, str(query))
        )

        found = query.one_or_none()
        found_fact = found.as_hamster(self.store) if found else None
        return found_fact

    # ***

    def strictly_during(self, since, until, result_limit=1000):
        """
        Return the fact(s) strictly contained within a since and until time.

        Args:
            since (datetime.datetime):
                Start datetime of facts to find.

            until (datetime.datetime):
                End datetime of facts to find.

            result_limit (int):
                Maximum number of facts to find, else log warning message.

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        query = self.store.session.query(AlchemyFact)

        condition = and_(
            func.datetime(AlchemyFact.start) >= query_prepare_datetime(since),
            or_(
                and_(
                    AlchemyFact.end != None,  # noqa: E711
                    func.datetime(AlchemyFact.end) <= query_prepare_datetime(until),
                ),
                and_(
                    AlchemyFact.end == None,  # noqa: E711
                    func.datetime(AlchemyFact.start) <= query_prepare_datetime(until),
                ),
            ),
        )

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        query = self.query_order_by_start(query, asc)

        self.store.logger.debug(
            "since: {} / until: {} / query: {}".format(since, until, str(query))
        )

        # LATER: (lb): We'll let the client ask for as many records as they
        # want. But we might want to offer ways to deal more gracefully with
        # it, like via pagination; or a fetch_one callback, so that only item
        # gets loaded in memory at a time, rather than everything. For now, we
        # can at least warn, I suppose.
        during_count = query.count()
        if during_count > result_limit:
            # (lb): hamster-lib would `raise OverflowError`,
            # but that seems drastic.
            message = _(
                "This is your alert that lots of Facts were found between "
                "the two dates specified: found {}.".factor(during_count)
            )
            self.store.logger.warning(message)

        facts = query.all()
        found_facts = [fact.as_hamster(self.store) for fact in facts]
        return found_facts

    # ***

    def surrounding(self, fact_time, inclusive=False):
        """
        Return the fact(s) at the given moment in time.
        Note that this excludes a fact that starts or ends at this time.
        (See antecedent and subsequent for finding those facts.)

        Args:
            fact_time (datetime.datetime):
                Time of fact(s) to match.

        Returns:
            list: List of ``nark.Facts`` instances.

        Raises:
            ValueError: If more than one Fact found at given time.
        """
        query = self.store.session.query(AlchemyFact)

        cmp_time = query_prepare_datetime(fact_time)

        if not inclusive:
            condition = and_(
                func.datetime(AlchemyFact.start) < cmp_time,
                # Find surrounding complete facts, or the ongoing fact.
                or_(
                    AlchemyFact.end == None,  # noqa: E711
                    func.datetime(AlchemyFact.end) > cmp_time,
                ),
            )
        else:
            condition = and_(
                func.datetime(AlchemyFact.start) <= cmp_time,
                # Find surrounding complete facts, or the ongoing fact.
                or_(
                    AlchemyFact.end == None,  # noqa: E711
                    func.datetime(AlchemyFact.end) >= cmp_time,
                ),
            )

        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        query = self.query_order_by_start(query, asc)

        self.store.logger.debug(
            "fact_time: {} / query: {}".format(fact_time, str(query))
        )

        if not inclusive:
            n_facts = query.count()
            if n_facts > 1:
                message = 'Broken time frame found at "{}": {} facts found'.format(
                    fact_time, n_facts
                )
                raise ValueError(message)

        facts = query.all()
        found_facts = [fact.as_hamster(self.store) for fact in facts]
        return found_facts

    # ***

    def endless(self):
        """
        Return any facts without a fact.start or fact.end.

        Args:
            <none>

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        query = self.store.session.query(AlchemyFact)

        # NOTE: (lb): Use ==/!=, not `is`/`not`, b/c SQLAlchemy
        #       overrides ==/!=, not `is`/`not`.
        condition = or_(
            AlchemyFact.start == None,  # noqa: E711
            AlchemyFact.end == None,  # noqa: E711
        )
        condition = and_(condition, AlchemyFact.deleted == False)  # noqa: E712

        query = query.filter(condition)

        self.store.logger.debug("query: {}".format(str(query)))

        facts = query.all()
        found_facts = [fact.as_hamster(self.store) for fact in facts]
        return found_facts

    # ***


# ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest

from nark.backends.sqlalchemy.objects import AlchemyCategory
from nark.items.category import Category


class TestCategoryManager:
    """"""

    def test_add_new(self, alchemy_store, alchemy_category_factory):
        """
        Our manager methods return the persistant instance as hamster objects.
        As we want to make sure that we compare our expectations against the
        raw saved object, we look it up explicitly again.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        category = alchemy_category_factory.build().as_hamster(alchemy_store)
        category.pk = None
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        result = alchemy_store.categories._add(category)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        db_instance = alchemy_store.session.query(AlchemyCategory).get(result.pk)
        assert category.equal_fields(db_instance)
        assert category != db_instance

    def test_add_existing_name(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that adding a category with a name that is already present
        gives an error.
        """
        existing_category = alchemy_category_factory()
        category = alchemy_category_factory.build().as_hamster(alchemy_store)
        category.name = existing_category.name
        category.pk = None
        with pytest.raises(ValueError):
            alchemy_store.categories._add(category)

    def test_add_with_pk(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that adding a alchemy_category that already got an PK
        raises an exception.
        """
        category = alchemy_category_factory().as_hamster(alchemy_store)
        category.name += "foobar"
        assert category.pk
        with pytest.raises(ValueError):
            alchemy_store.categories._add(category)

    def test_update(
        self,
        alchemy_store,
        alchemy_category_factory,
        new_category_values,
    ):
        """Test that updateing a alchemy_category works as expected."""
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        # (lb): NOTE_TO_SELF: The alchemy_category_factory fixture was created
        # when conftest.py called `register(lib_factories.CategoryFactory)`,
        # which, AFAICT, is the same as calling the factory directly:
        #   alchemy_category_factory = factories.AlchemyCategoryFactory
        category = alchemy_category_factory().as_hamster(alchemy_store)
        new_values = new_category_values(category)
        for key, value in new_values.items():
            assert getattr(category, key) != value
        for key, value in new_values.items():
            setattr(category, key, value)
        alchemy_store.categories._update(category)
        db_instance = alchemy_store.session.query(AlchemyCategory).get(category.pk)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert category.equal_fields(db_instance)

    def test_update_without_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure that passing a category without a PK raises an error."""
        category = alchemy_category_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.categories._update(category)

    def test_update_invalid_pk(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that passing a category with a non existing PK raises an error.
        """
        category = alchemy_category_factory().as_hamster(alchemy_store)
        category.pk = category.pk + 10
        with pytest.raises(KeyError):
            alchemy_store.categories._update(category)

    def test_update_existing_name(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that renaming a given alchemy_category to a taken name
        throws an error.
        """
        category_1, category_2 = (
            alchemy_category_factory(),
            alchemy_category_factory(),
        )
        category_2 = category_2.as_hamster(alchemy_store)
        category_2.name = category_1.name
        with pytest.raises(ValueError):
            alchemy_store.categories._update(category_2)

    def test_remove(self, alchemy_store, alchemy_category_factory):
        """Make sure passing a valid alchemy_category removes it from the db."""
        alchemy_category = alchemy_category_factory()
        category = alchemy_category.as_hamster(alchemy_store)
        alchemy_store.categories.remove(category)
        assert alchemy_store.session.query(AlchemyCategory).get(category.pk) is None

    def test_remove_no_pk(self, alchemy_store, alchemy_category_factory):
        """Ensure that passing a alchemy_category without an PK raises an error."""
        category = alchemy_category_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.categories.remove(category)

    def test_remove_invalid_pk(self, alchemy_store, alchemy_category_factory):
        """Ensure that passing a alchemy_category without an PK raises an error."""
        category = alchemy_category_factory.build(pk=800).as_hamster(alchemy_store)
        with pytest.raises(KeyError):
            alchemy_store.categories.remove(category)

    def test_get_existing_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure method retrieves corresponding object."""
        category = alchemy_category_factory().as_hamster(alchemy_store)
        result = alchemy_store.categories.get(category.pk)
        assert result == category

    def test_get_non_existing_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure we throw an error if PK can not be resolved."""
        alchemy_store.session.query(AlchemyCategory).count == 0
        category = alchemy_category_factory()
        alchemy_store.session.query(AlchemyCategory).count == 1
        with pytest.raises(KeyError):
            alchemy_store.categories.get(category.pk + 1)

    def test_get_by_name(self, alchemy_category_factory, alchemy_store):
        """Make sure a alchemy_category can be retrieved by name."""
        category = alchemy_category_factory().as_hamster(alchemy_store)
        result = alchemy_store.categories.get_by_name(category.name)
        assert result == category

    def test_get_all(self, alchemy_store, set_of_categories):
        results = alchemy_store.categories.get_all()
        assert len(results) == len(set_of_categories)
        assert len(results) == alchemy_store.session.query(AlchemyCategory).count()
        for category in set_of_categories:
            assert category.as_hamster(alchemy_store) in results

    # Test convenience methods.
    def test_get_or_create_get(self, alchemy_store, alchemy_category_factory):
        """
        Test that if we pass a alchemy_category of existing name, we just
        return it.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        category = alchemy_category_factory().as_hamster(alchemy_store)
        result = alchemy_store.categories.get_or_create(category)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert result == category

    def test_get_or_create_new_name(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that passing a category with new name creates and returns
        new instance.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        category = alchemy_category_factory.build().as_hamster(alchemy_store)
        category.pk = None
        result = alchemy_store.categories.get_or_create(category)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert result.equal_fields(category)

    def test_get_deleted_item(self, alchemy_store, alchemy_category):
        """Make sure method retrieves deleted object."""
        alchemy_category.deleted = True
        result = alchemy_store.categories.get(alchemy_category.pk, deleted=True)
        assert result == alchemy_category

    def test_get_all_match_categories(
        self,
        alchemy_store,
        set_of_alchemy_facts_active,
        alchemy_category_factory,
    ):
        """Test get_all argument: QueryTerms.match_activities."""
        category_0 = set_of_alchemy_facts_active[0].activity.category
        category_2 = set_of_alchemy_facts_active[2].activity.category.name
        any_category = alchemy_category_factory(pk=None)
        category_with_pk_None = Category(name=any_category.name)
        results = alchemy_store.categories.get_all(
            match_categories=[
                category_0,
                category_2,
                category_with_pk_None,
                None,
            ],
        )
        assert len(results) == 3

    # ***

    @pytest.mark.parametrize(
        ("sort_cols"),
        (
            (["start"]),
            ([None]),
            (["usage"]),
            (["time"]),
            (["activity"]),
            (["category"]),
        ),
    )
    def test_get_all_sort_cols(self, alchemy_store, sort_cols):
        alchemy_store.categories.get_all(sort_cols=sort_cols)

    def test_get_all_sort_cols_tag_fails(self, alchemy_store, mocker):
        # The tag table is not joined for Activity query.
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.categories.get_all(sort_cols=["tag"])
        assert alchemy_store.logger.warning.called

    def test_get_all_sort_cols_unknown(self, alchemy_store, mocker):
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.categories.get_all(sort_cols=["foo"])
        assert alchemy_store.logger.warning.called



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import copy
import datetime
from math import inf
from operator import attrgetter

import faker as faker_
import pytest
from freezegun import freeze_time

from nark.items.activity import Activity
from nark.items.category import Category
from nark.items.fact import Fact
from nark.items.tag import Tag

from .test_activity import TestActivity
from .test_tag import TestTag

faker = faker_.Faker()


# A class to test as_kvals.
class FactWithFact(Fact):
    def __init__(self, *args, **kwargs):
        super(FactWithFact, self).__init__(*args, **kwargs)
        self.mirror = self


class TestFact(object):
    """items.fact.Fact class tests."""

    def test_fact_init_valid(
        self,
        activity,
        start_end_datetimes,
        pk_valid_parametrized,
        description_valid_parametrized,
        tag_list_valid_parametrized,
    ):
        """Make sure valid values instaniate a Fact."""
        fact = Fact(
            activity,
            start_end_datetimes[0],
            start_end_datetimes[1],
            pk_valid_parametrized,
            description_valid_parametrized,
            tag_list_valid_parametrized,
        )
        assert fact.activity == activity
        assert fact.pk == pk_valid_parametrized
        assert fact.description == description_valid_parametrized
        assert fact.start == start_end_datetimes[0]
        assert fact.end == start_end_datetimes[1]
        # tag_list_valid_parametrized is a set() of name strings.
        names = list(tag_list_valid_parametrized)
        tags = set([Tag(pk=None, name=name) for name in names])
        tags = sorted(list(tags), key=attrgetter("name"))
        assert fact.tags_sorted == tags

    # ***

    @pytest.mark.parametrize(
        ("factoid", "time_hint", "lenient", "should_err"),
        [
            (
                "12:00 - 14:00 foo@bar, bazbat",
                "verify_both",
                False,
                None,
            ),
            (
                "12:00 - 14:00 foo",
                "verify_both",
                True,
                "Expected to find an Activity name.",
            ),
            (
                "foo@bar",
                "verify_none",
                True,
                None,
            ),
            (
                "12:00-14:00 foo@bar",
                "verify_both",
                False,
                None,
            ),
            # Test seconds (2018-08-16: they no longer get truncated).
            (
                "12:00:11 - 14:00:59 baz@bat",
                "verify_both",
                False,
                None,
            ),
            # Test just start and end, no activity, category, tags, nor description.
            (
                "12:00:11 - 13:01",
                "verify_both",
                True,
                "Expected to find an Activity name.",
            ),
            # Test just a start time.
            (
                "13:01:22",
                "verify_start",
                True,
                "Expected to find an Activity name.",
            ),
        ],
    )
    def test_create_from_factoid_valid(self, factoid, time_hint, lenient, should_err):
        """Make sure that a valid raw fact creates a proper Fact."""
        fact, err = Fact.create_from_factoid(
            factoid,
            time_hint=time_hint,
            lenient=lenient,
        )
        assert fact
        assert str(err) == str(should_err)

    @pytest.mark.parametrize(
        ("raw_fact", "expectations"),
        [
            (
                # Note that without a time_hint, no time expected, so
                # start and end are None, and prefix (before colon) is
                # just part of the activity name.
                "-7 foo@bar, bazbat",
                {
                    "start": None,
                    "end": None,
                    "activity": "-7 foo",
                    "category": "bar",
                    "description": "bazbat",
                },
            ),
        ],
    )
    @freeze_time("2015-05-02 18:07")
    def test_create_from_factoid_with_delta_no_time_hint(self, raw_fact, expectations):
        fact, err = Fact.create_from_factoid(raw_fact)
        assert fact.start == expectations["start"]
        assert fact.end == expectations["end"]
        assert fact.activity.name == expectations["activity"]
        assert fact.activity.category.name == expectations["category"]
        assert fact.description == expectations["description"]
        assert not err

    @pytest.mark.parametrize(
        ("factoid", "expectations"),
        [
            (
                "-7 foo@bar, bazbat",
                {
                    # Note that time parsing does not occur until later, so that
                    # multiple Factoids can be parsed, and then relative times can
                    # be determined with reference to surrounding Facts. Although in
                    # this example, because start is relative to end, we could add code
                    # to resolve time relative to the Fact itself, but there's not a
                    # good reason to do so, e.g., here, after resolution, you'd see:
                    #   'start': datetime.datetime(2015, 5, 2, 18, 0, 0),
                    "start": "-7",
                    "end": None,
                    "activity": "foo",
                    "category": "bar",
                    "description": "bazbat",
                },
            ),
        ],
    )
    @freeze_time("2015-05-02 18:07")
    def test_create_from_factoid_with_delta_time_hint_start(
        self, factoid, expectations
    ):
        fact, err = Fact.create_from_factoid(factoid, time_hint="verify_start")
        assert fact.start == expectations["start"]
        assert fact.end == expectations["end"]
        assert fact.activity.name == expectations["activity"]
        assert fact.activity.category.name == expectations["category"]
        assert fact.description == expectations["description"]
        assert not err

    # ***

    @pytest.mark.parametrize(
        "start",
        [
            None,
            # The faker randomizes the microsecond field, but nark assumes 0.
            faker.date_time().replace(microsecond=0),
            "10",
            "+10",
            "-10h5m",
        ],
    )
    def test_start_valid(self, fact, start):
        """Make sure that valid arguments get stored by the setter."""
        fact.start = start
        assert fact.start == start

    @pytest.mark.parametrize(
        "start",
        [
            faker.date_time().strftime("%y-%m-%d %H:%M"),  # string, not datetime
            "not relative",
            "+10d",  # Not supported
        ],
    )
    def test_start_invalid(self, fact, start):
        """Make sure that trying to store dateimes as strings throws an error."""
        with pytest.raises(TypeError):
            fact.start = start

    @pytest.mark.parametrize(
        "end",
        [
            None,
            # The faker randomizes the microsecond field, but nark assumes 0.
            faker.date_time().replace(microsecond=0),
        ],
    )
    def test_end_valid(self, fact, end):
        """Make sure that valid arguments get stored by the setter."""
        fact.end = end
        assert fact.end == end

    def test_end_invalid(self, fact):
        """Make sure that trying to store dateimes as strings throws an error."""
        with pytest.raises(TypeError):
            # No need to replace microseconds, because %M truncates (m)secs.
            fact.end = faker.date_time().strftime("%y-%m-%d %H:%M")

    def test_description_valid(self, fact, description_valid_parametrized):
        """Make sure that valid arguments get stored by the setter."""
        fact.description = description_valid_parametrized
        assert fact.description == description_valid_parametrized

    def test_category_property(self, fact):
        """Make sure the property returns this facts category."""
        assert fact.category == fact.activity.category

    def test_serialized_string(self, fact):
        """
        Ensure a serialized string with full information matches our expectation.
        """
        expect_f = "{start} to {end} {activity}@{category}: #{tag}: {description}"
        expectation = expect_f.format(
            start=fact.start.strftime("%Y-%m-%d %H:%M:%S"),
            end=fact.end.strftime("%Y-%m-%d %H:%M:%S"),
            activity=fact.activity.name,
            category=fact.category.name,
            tag=sorted(list(fact.tags), key=attrgetter("name"))[0].name,
            description=fact.description,
        )
        result = fact.get_serialized_string()
        assert isinstance(result, str)
        assert result == expectation

    @pytest.mark.parametrize(
        ("values", "expectation"),
        (
            (
                {
                    "start": datetime.datetime(2016, 1, 1, 18),
                    "end": datetime.datetime(2016, 1, 1, 19),
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set([Tag("math"), Tag("science")]),
                    "description": "something clever ...",
                },
                "2016-01-01 18:00:00 to 2016-01-01 19:00:00 "
                "homework@school: #math #science: something clever ...",
            ),
            (
                {
                    "start": datetime.datetime(2016, 1, 1, 18),
                    "end": datetime.datetime(2016, 1, 1, 19),
                    "activity": Activity("homework", category=None),
                    "tags": set([Tag("math"), Tag("science"), Tag("science fiction")]),
                    "description": "something",
                },
                "2016-01-01 18:00:00 to 2016-01-01 19:00:00 "
                "homework@: #math #science #science fiction: something",
            ),
            (
                {
                    "start": datetime.datetime(2016, 1, 1, 18),
                    "end": datetime.datetime(2016, 1, 1, 19),
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set(),
                    "description": "something clever ...",
                },
                "2016-01-01 18:00:00 to 2016-01-01 19:00:00 "
                "homework@school: something clever ...",
            ),
            (
                {
                    "start": datetime.datetime(2016, 1, 1, 18),
                    "end": datetime.datetime(2016, 1, 1, 19),
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set([Tag("science"), Tag("math")]),
                    "description": "",
                },
                "2016-01-01 18:00:00 to 2016-01-01 19:00:00 "
                "homework@school: #math #science",
            ),
            (
                {
                    "start": datetime.datetime(2016, 1, 1, 18),
                    "end": datetime.datetime(2016, 1, 1, 19),
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set(),
                    "description": "",
                },
                "2016-01-01 18:00:00 to 2016-01-01 19:00:00 " "homework@school",
            ),
            (
                {
                    "start": None,
                    "end": datetime.datetime(2016, 1, 1, 19),
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set([Tag("math"), Tag("science")]),
                    "description": "something clever ...",
                },
                # FIXME/2018-08-17 12:25: Update factoid parse to recognize
                # 'to' and 'at' prefix to distinguish between verify_end, verify_start?
                # and then anything else is verify_both or verify_none?? hrmmmm...
                # maybe the answer is a 2nd parse-factoid wrapper, i.e.,
                #   one parser for verify_hint, and one parser for unknown-hint...
                "2016-01-01 19:00:00 homework@school: #math #science: something clever ...",  # noqa: E501
            ),
            (
                {
                    "start": None,
                    "end": None,
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set([Tag("math"), Tag("science")]),
                    "description": "something clever ...",
                },
                # FIXME: Make new parse wrapper that checks for 'to' 'at', or date.
                #   Then look for 'to <date>', 'at <date>', etc.
                #   Fall back to what? Expect no dates? Both dates?
                #   Problem really is that I feel a Fact with no start or end
                #     is really an invalid Fact!! So this Factoid should be a
                #     problem, right?:
                "homework@school: #math #science: something clever ...",
            ),
            (
                {
                    "start": datetime.datetime(2016, 1, 1, 18),
                    "end": None,
                    "activity": Activity("homework", category=Category("school")),
                    "tags": set([Tag("math"), Tag("science")]),
                    "description": "something clever ...",
                },
                "at 2016-01-01 18:00:00 "
                "homework@school: #math #science: something clever ...",
            ),
        ),
    )
    def test_serialized_string_various_missing_values(self, fact, values, expectation):
        """
        Make sure the serialized string is correct even if some information is missing.
        """
        for attribute, value in values.items():
            setattr(fact, attribute, value)
        assert fact.get_serialized_string() == expectation

    def test__eq__false(self, fact):
        """Make sure that two distinct facts return ``False``."""
        other = copy.deepcopy(fact)
        other.pk = 1
        assert fact is not other
        assert fact != other

    def test__eq__true(self, fact):
        """Make sure that two identical facts return ``True``."""
        other = copy.deepcopy(fact)
        assert fact is not other
        assert fact == other

    def test_is_hashable(self, fact):
        """Test that ``Fact`` instances are hashable."""
        assert hash(fact)

    def test_hash_method(self, fact):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(fact) == hash(fact.as_tuple())

    def test_hash_different_between_instances(self, fact_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Fact.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(fact_factory()) != hash(fact_factory())

    def test__gt__(self, fact):
        # A little self-referential, but covers it!
        assert not (fact > fact)

    def test__lt__(self, fact):
        # A little self-referential, but covers it!
        assert not (fact < fact)

    def test_sorty_times(self, fact):
        assert fact.sorty_times == (fact.start, fact.end)

    def test_sorty_tuple_inf(self, fact):
        assert fact.pk is None  # Which sorty_tuple replaces with -inf.
        assert fact.sorty_tuple == (fact.start, fact.end, -inf)

    def test_sorty_tuple_pk(self, fact):
        fact.pk = 123
        assert fact.sorty_tuple == (fact.start, fact.end, fact.pk)

    def test_as_kvals_avoid_repr_recursion(self):
        fact = FactWithFact(activity=None, start=None)
        repred = fact.as_kvals()
        assert repred.startswith("Fact(")

    def test_copy_include_pk(self, fact):
        new_fact = fact.copy(include_pk=True)
        assert new_fact == fact

    def test_as_tuple_include_pk(self, fact):
        """Make sure that conversion to a tuple matches our expectations."""
        assert fact.as_tuple() == (
            fact.pk,
            fact.activity.as_tuple(include_pk=True),
            fact.start,
            fact.end,
            fact.description,
            frozenset(fact.tags),
            fact.deleted,
            fact.split_from,
        )

    def test_as_tuple_exclude_pk(self, fact):
        """Make sure that conversion to a tuple matches our expectations."""
        assert fact.as_tuple(include_pk=False) == (
            False,
            fact.activity.as_tuple(include_pk=False),
            fact.start,
            fact.end,
            fact.description,
            frozenset([tag.as_tuple(include_pk=False) for tag in fact.tags]),
            fact.deleted,
            fact.split_from,
        )

    def test_equal_fields_true(self, fact):
        """Make sure that two facts that differ only in their PK compare equal."""
        other = copy.deepcopy(fact)
        other.pk = 1
        assert fact.equal_fields(other)

    def test_equal_fields_false(self, fact):
        """Make sure that two facts that differ not only in their PK compare unequal."""
        other = copy.deepcopy(fact)
        other.pk = 1
        other.description += "foobar"
        assert fact.equal_fields(other) is False

    # ***

    def test_start_fmt(self, fact):
        assert fact.start_fmt() == fact.start.strftime("%Y-%m-%d %H:%M:%S")

    def test_start_fmt_utc_empty(self):
        fact = Fact(activity=None, start=None)
        assert fact.start_fmt_utc == ""

    def test_start_fmt_utc_valid(self, fact):
        formatted = fact.start_fmt_utc
        assert formatted == fact.start.strftime("%Y-%m-%d %H:%M:%S%z")

    def test_start_fmt_local_empty(self):
        fact = Fact(activity=None, start=None)
        assert fact.start_fmt_local == ""

    def test_start_fmt_local_valid(self, fact):
        formatted = fact.start_fmt_local
        assert formatted == fact.start.strftime("%Y-%m-%d %H:%M:%S%z")

    # ***

    def test_end_fmt(self, fact):
        assert fact.end_fmt() == fact.end.strftime("%Y-%m-%d %H:%M:%S")

    def test_end_fmt_utc_empty(self):
        fact = Fact(activity=None, start=None, end=None)
        assert fact.end_fmt_utc == ""

    def test_end_fmt_utc_valid(self, fact):
        formatted = fact.end_fmt_utc
        assert formatted == fact.end.strftime("%Y-%m-%d %H:%M:%S%z")

    def test_end_fmt_local_empty(self):
        fact = Fact(activity=None, start=None, end=None)
        assert fact.end_fmt_local == ""

    def test_end_fmt_local_valid(self, fact):
        formatted = fact.end_fmt_local
        assert formatted == fact.end.strftime("%Y-%m-%d %H:%M:%S%z")

    @freeze_time("2015-05-02 18:07")
    def test_end_fmt_local_nowwed_now(self):
        fact = Fact(activity=None, start=None, end=None)
        formatted = fact.end_fmt_local_nowwed
        assert formatted == "2015-05-02 18:07:00 <now>"

    @freeze_time("2015-05-02 18:07")
    def test_end_fmt_local_nowwed_end(self, fact):
        formatted = fact.end_fmt_local_nowwed
        assert formatted == fact.end.strftime("%Y-%m-%d %H:%M:%S")

    #    @freeze_time('2015-05-02 18:07')
    #    def test_end_fmt_local_nowwed_now(self):
    #        fact = Fact(activity=None, start=None, end=None)
    #        formatted = fact.end_fmt_local_nowwed
    #        assert formatted == '2015-05-02 18:07:00 <now>'

    @freeze_time("2015-05-02 18:07")
    def test_end_fmt_local_or_now_end(self, fact):
        formatted = fact.end_fmt_local_or_now
        assert formatted == fact.end.strftime("%Y-%m-%d %H:%M:%S")

    @freeze_time("2015-05-02 18:07")
    def test_end_fmt_local_or_now_now(self):
        fact = Fact(activity=None, start=None, end=None)
        formatted = fact.end_fmt_local_or_now
        assert formatted == "2015-05-02 18:07:00"

    # ***

    def test_momentaneous_False(self, fact):
        assert not fact.momentaneous

    def test_momentaneous_True(self, fact):
        fact.end = fact.start
        assert fact.momentaneous

    @freeze_time("2015-05-02 18:07")
    def test_time_now(self, fact):
        # (lb): What's the point of accessing 'now' through the Fact? I forget.
        assert fact.time_now == datetime.datetime.now()

    def test_times(self, fact):
        # Note that Fact.times is similar to Fact.sorty_times, except latter
        # replaces Fact.end with UntilTimeStops if fact.end is None.
        assert fact.times == (fact.start, fact.end)

    def test_times_ok_True(self, fact):
        assert fact.times_ok is True

    def test_times_ok_False(self):
        fact = Fact(activity=None, start=None)
        assert fact.times_ok is False

    # ***

    def test_delta(self, fact):
        """Make sure that valid arguments get stored by the setter."""
        assert fact.delta() == fact.end - fact.start

    @freeze_time("2015-05-02 18:07")
    def test_delta_no_end(self, fact):
        """Make sure that a missing end datetime results in ``delta=None``."""
        # See FactFactory for default start/end.
        fact.end = None
        # NOTE: With freezegun, both now() and utcnow() are the same.
        assert fact.delta() == (datetime.datetime.now() - fact.start)

    @pytest.mark.parametrize(
        "offset",
        [
            (
                15,
                {
                    "%M": "15",
                    "%H:%M": "00:15",
                    "HHhMMm": " 0 hours 15 minutes",
                    "": "15.00 mins.",
                },
            ),
            (
                452,
                {
                    "%M": "452",
                    "%H:%M": "07:32",
                    "HHhMMm": " 7 hours 32 minutes",
                    "": "7.53 hours",
                },
            ),
            (
                912,
                {
                    "%M": "912",
                    "%H:%M": "15:12",
                    "HHhMMm": "15 hours 12 minutes",
                    "": "15.20 hours",
                },
            ),
            (
                61,
                {
                    "%M": "61",
                    "%H:%M": "01:01",
                    "HHhMMm": " 1 hour   1 minute ",
                    "": "1.02 hours",
                },
            ),
        ],
    )
    def test_format_delta_valid_style(
        self,
        fact,
        offset,
        start_end_datetimes_from_offset_now,
        string_delta_style_parametrized,
    ):
        """Make sure that the resulting string matches our expectation."""
        end_offset, expectation = offset
        fact.start, fact.end = start_end_datetimes_from_offset_now(end_offset)
        result = fact.format_delta(style=string_delta_style_parametrized)
        assert result == expectation[string_delta_style_parametrized]

    def test_format_delta_invalid_style(self, fact):
        """Ensure that passing an invalid format will raise an exception."""
        with pytest.raises(ValueError):
            fact.format_delta(style="foobar")

    # ***

    def test_midpoint_invalid(self):
        fact = Fact(activity=None, start=None)
        assert fact.midpoint is None

    def test_midpoint_valid(self, fact):
        assert fact.midpoint == fact.end - ((fact.end - fact.start) / 2)

    def test_time_of_day_midpoint_invalid(self):
        fact = Fact(activity=None, start=None)
        assert fact.time_of_day_midpoint == ""

    def test_time_of_day_midpoint_valid(self, fact):
        todm = fact.time_of_day_midpoint
        assert " ◐ " in todm

    def test_time_of_day_humanize_invalid_no_start_no_end(self):
        start = datetime.datetime(2015, 12, 10, 12, 30, 33)
        fact = Fact(activity=None, start=start, end=None)
        assert fact.time_of_day_humanize() == ""

    @freeze_time("2015-12-10 18:07")
    def test_time_of_day_humanize_invalid_no_start_no_end_show_now(self):
        start = datetime.datetime(2015, 12, 10, 12, 30, 33)
        fact = Fact(activity=None, start=start, end=None)
        todh = fact.time_of_day_humanize(show_now=True)
        assert todh == "Thu 10 Dec 2015 ◐ 12:30 PM — 06:07 PM"

    def test_time_of_day_humanize_momentaneous(self):
        a_moment = datetime.datetime(1974, 5, 22, 8, 24, 18)
        fact = Fact(activity=None, start=a_moment, end=a_moment)
        todh = fact.time_of_day_humanize()
        assert todh == "Wed 22 May 1974 ◐ 08:24 AM"

    def test_time_of_day_humanize_same_day(self):
        start = datetime.datetime(2015, 12, 10, 12, 30, 33)
        end = datetime.datetime(2015, 12, 10, 13, 30, 33)
        fact = Fact(activity=None, start=start, end=end)
        todh = fact.time_of_day_humanize()
        assert todh == "Thu 10 Dec 2015 ◐ 12:30 PM — 01:30 PM"

    def test_time_of_day_humanize_separate_days(self):
        start = datetime.datetime(2015, 12, 10, 12, 30, 33)
        end = datetime.datetime(2015, 12, 12, 13, 30, 33)
        fact = Fact(activity=None, start=start, end=end)
        todh = fact.time_of_day_humanize()
        assert todh == "Thu 10 Dec 2015 ◐ 12:30 PM — 01:30 PM Sat 12 Dec 2015"

    # ***

    def test_fact_activity_name_invalid(self, fact):
        fact.activity = None
        assert fact.activity_name == ""

    def test_fact_activity_name_valid(self, fact):
        assert fact.activity_name == fact.activity.name

    def test_fact_category(self, fact):
        assert fact.category is fact.activity.category

    def test_fact_category_name_invalid(self, fact):
        fact.activity = None
        assert fact.category_name == ""

    def test_fact_category_name_valid(self, fact):
        assert fact.category_name == fact.activity.category.name

    # ***

    # (lb): Skipping fact.description* and fact.tag* methods that start
    #       at and follow fact.description() (in the module, as of 2020-06-20)
    #       because I see coverage provided incidentally (from other tests).
    #       (And my goal at this moment is 100% coverage, but not necessarily
    #       from this specific test class.)

    # ***

    def test_oid_actegory_empty_default(self):
        fact = Fact(activity=None, start=None)
        assert fact.oid_actegory() == "@"

    def test_oid_actegory_empty_supplied(self):
        fact = Fact(activity=None, start=None)
        empty_actegory_placeholder = "foo"
        txt = fact.oid_actegory(empty_actegory_placeholder=empty_actegory_placeholder)
        assert txt == empty_actegory_placeholder

    def test_oid_actegory_nameless_activity(self, fact):
        fact.activity.name = ""
        assert fact.oid_actegory() == "@{}".format(fact.category_name)

    def test_oid_actegory_both_named_unshelled(self, fact):
        expected = "{}@{}".format(fact.activity_name, fact.category_name)
        assert fact.oid_actegory() == expected

    def test_oid_actegory_both_named_shellify(self, fact):
        expected = '"{}@{}"'.format(fact.activity_name, fact.category_name)
        assert fact.oid_actegory(shellify=True) == expected

    # ***

    def test_oid_description_cut_width(self, fact):
        fact.description = "abc de fghijkl"
        txt = fact.oid_description(cut_width=9, sep=", ")
        assert txt == ", abc de..."

    # ***

    def test_oid_tags_quote_tokens(self, fact):
        txt = fact.oid_tags(quote_tokens=True)
        assert txt.startswith('"#')

    # ***

    def test_friendly_str_basic(self, fact):
        txt = fact.friendly_str()
        assert txt

    def test_friendly_str_localize_True(self, fact):
        Fact.localize(localize=True)
        txt = fact.friendly_str()
        Fact.localize(localize=False)
        # (lb): Being lazy. This test provides coverage, does not check
        # that the start and end time are localized.
        assert txt

    def test_friendly_str_include_id(self, fact):
        txt = fact.friendly_str(include_id=True)
        assert fact.pk is None
        assert txt.startswith("(🏭 None) ")

    def test_friendly_str_cut_width(self):
        start = datetime.datetime(2015, 12, 10, 12, 30, 33)
        fact = Fact(activity=None, start=start, end=None)
        txt = fact.friendly_str(cut_width_complete=16)
        assert txt == "at 2015-12-10..."

    def test_friendly_str_show_elapsed(self, fact):
        txt = fact.friendly_str(show_elapsed=True)
        # (lb): Being lazy. This test provides coverage, does not check
        # for the elapsed time, e.g., ' [3.00 hours] ' in txt.
        assert txt

    # ***

    def test_fact_short(self, fact):
        assert fact.short

    # ***

    def test_create_from_parsed_fact_no_activity_raises(self, fact):
        parsed_fact = {
            "start": fact.start,
            "end": fact.end,
            "activity": None,
        }
        with pytest.raises(ValueError):
            Fact.create_from_parsed_fact(parsed_fact)

    # ***

    def test__str__(self, fact):
        expect_f = "{start} to {end} {activity}@{category}: {tags}: {description}"
        expectation = expect_f.format(
            start=fact.start.strftime("%Y-%m-%d %H:%M:%S"),
            end=fact.end.strftime("%Y-%m-%d %H:%M:%S"),
            activity=fact.activity.name,
            category=fact.category.name,
            tags=fact.tagnames(),
            description=fact.description,
        )
        assert str(fact) == expectation

    def test__str__no_end(self, fact):
        fact.end = None
        expect_f = "at {start} {activity}@{category}: {tags}: {description}"
        expectation = expect_f.format(
            start=fact.start.strftime("%Y-%m-%d %H:%M:%S"),
            activity=fact.activity.name,
            category=fact.category.name,
            tags=fact.tagnames(),
            description=fact.description,
        )
        assert str(fact) == expectation

    def test__str__no_start_no_end(self, fact):
        fact.start = None
        fact.end = None
        expectation = "{activity}@{category}: {tags}: {description}".format(
            activity=fact.activity.name,
            category=fact.category.name,
            tags=fact.tagnames(),
            description=fact.description,
        )
        assert str(fact) == expectation

    # (lb): It might be nice to do snapshot testing. However, that won't
    # work unless we disable the random string generator we use to make up
    # item names.
    #
    # Here's what a snapshot test might look like:
    #
    #   def test__repr__snapshot(self, snapshot, fact):
    #       """
    #       Test repr() against snapshot. Save time not writing test expectation.
    #       """
    #       result = repr(fact)
    #       snapshot.assert_match(result)
    #
    # In lieu of that, we re-generate the repr() herein. Which makes me
    # feel weird, like we're just duplicating what Fact.__repr__ already
    # does.

    def assert_fact_repr(self, the_fact):
        # (lb): I feel somewhat dirty with the test__repr__ methods.
        #   I feel like these should be snapshot tests, and not tests
        #   that require manual labor to maintain a tedious string
        #   builder that basically mimics the behavior of the methods
        #   that we're testing. Blech.
        tag_parts = []
        for tag in the_fact.tags:
            tag_parts.append(TestTag.as_repr(tag))
        tags = ", ".join(tag_parts)
        expect_f = (
            "Fact("
            "_description={description}, "
            "_end={end}, "
            "_start={start}, "
            "activity={activity}, "
            "deleted={deleted}, "
            "pk={pk}, "
            "split_from={split_from}, "
            "tags=[{tags}]"
            ")"
        )
        expectation = expect_f.format(
            pk=repr(the_fact.pk),
            split_from=repr(the_fact.split_from),
            start=repr(the_fact.start),
            end=repr(the_fact.end),
            activity=TestActivity.as_repr(the_fact.activity),
            tags=tags,
            description=repr(the_fact.description),
            deleted=repr(the_fact.deleted),
        )
        result = repr(the_fact)
        assert isinstance(result, str)
        assert result == expectation

    def test__repr__(self, fact):
        """Make sure our debugging representation matches our expectations."""
        self.assert_fact_repr(fact)

    def test__repr__no_end(self, fact):
        """Test that facts without end datetime are represented properly."""
        result = repr(fact)
        assert isinstance(result, str)
        fact.end = None
        self.assert_fact_repr(fact)

    def test__repr__no_start_no_end(self, fact):
        """Test that facts without times are represented properly."""
        fact.start = None
        fact.end = None
        self.assert_fact_repr(fact)

    def test__str__no_tags(self, fact):
        fact.tags = []
        self.assert_fact_repr(fact)



#!/usr/bin/env python

# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from sqlalchemy_migrate_hotoffthehamster.versioning.shell import main

if __name__ == "__main__":
    main(debug="False")



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest
from freezegun import freeze_time

from nark.backends.sqlalchemy.objects import AlchemyActivity, AlchemyFact, AlchemyTag


class TestFactManager:
    """"""

    def test_add_fails_timeframe_available_existing_fact_overlaps_start_only(
        self,
        alchemy_store,
        fact,
        alchemy_fact,
    ):
        """
        Make sure that passing a fact with only start overlapping an existing
        one raises error.
        """
        fact.start = alchemy_fact.start - datetime.timedelta(days=4)
        fact.end = alchemy_fact.start + datetime.timedelta(minutes=15)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    def test_add_fails_timeframe_available_existing_fact_overlaps_end_only(
        self,
        alchemy_store,
        fact,
        alchemy_fact,
    ):
        """
        Make sure that passing a fact with only end overlapping an existing
        one raises error.
        """
        fact.start = alchemy_fact.end - datetime.timedelta(minutes=1)
        fact.end = alchemy_fact.end + datetime.timedelta(minutes=15)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    # Testcase for Bug LIB-253
    def test_add_fails_timeframe_available_fact_within_existing_timeframe(
        self,
        alchemy_store,
        fact,
        alchemy_fact,
    ):
        """
        Make sure that passing a fact that is completely within an existing
        ones raises error.
        """
        fact.start = alchemy_fact.start + datetime.timedelta(minutes=1)
        fact.end = alchemy_fact.end - datetime.timedelta(minutes=1)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    def test_add_fails_timeframe_available_fact_spans_existing_timeframe(
        self,
        alchemy_store,
        fact,
        alchemy_fact,
    ):
        """
        Make sure that passing a fact that completely spans an existing fact
        raises an error.
        """
        fact.start = alchemy_fact.start - datetime.timedelta(minutes=1)
        fact.end = alchemy_fact.end + datetime.timedelta(minutes=15)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    # ***

    def test_add_tags(self, alchemy_store, fact):
        """Make sure that adding a new valid fact will also save its tags."""
        result = alchemy_store.facts._add(fact)
        assert fact.tags
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert db_instance.tags
        assert db_instance.as_hamster(alchemy_store).equal_fields(fact)

    def test_add_new_valid_fact_new_activity(self, alchemy_store, fact):
        """
        Make sure that adding a new valid fact with a new activity works as
        intended.
        """
        assert alchemy_store.session.query(AlchemyFact).count() == 0
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        result = alchemy_store.facts._add(fact)
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert alchemy_store.session.query(AlchemyFact).count() == 1
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(fact)

    def test_add_new_valid_fact_existing_activity(
        self,
        alchemy_store,
        fact,
        alchemy_activity,
    ):
        """
        Make sure that adding a new valid fact with an existing activity works
        as intended.
        """
        fact.activity = alchemy_activity.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyFact).count() == 0
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        result = alchemy_store.facts._add(fact)
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert alchemy_store.session.query(AlchemyFact).count() == 1
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(fact)

    def test_add_fails_pk_not_none(self, alchemy_store, fact):
        """Make sure that passing a fact with a PK raises error."""
        fact.pk = 101
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    # ***

    def test_update_respects_tags(self, alchemy_store, alchemy_fact, new_fact_values):
        """Make sure that updating sets tags as expected."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        new_values = new_fact_values(fact)
        fact.tags = new_values["tags"]
        result = alchemy_store.facts._update(fact)
        # Because split_from, fact will have been marked deleted.
        assert fact.deleted
        # And the new Fact will have a larger PK.
        # (lb): We really don't need to impose a strictly increasing order
        # property on the PK, but it's how our code happens to work, so might
        # as well check it.
        assert result.pk > fact.pk
        fact.deleted = False
        assert result.split_from.pk == fact.pk
        # Note that the split-from is not the same because it contains previous tags.
        assert result.split_from != fact
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert db_instance.as_hamster(alchemy_store).equal_fields(result)
        result.split_from = None
        assert result.equal_fields(fact)

    def test_update_fails_pk_unknown(
        self, alchemy_store, alchemy_fact, new_fact_values
    ):
        """Make sure that trying to update a fact that does not exist raises error."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        new_values = new_fact_values(fact)
        fact.start = new_values["start"]
        fact.end = new_values["end"]
        fact.pk += 100
        with pytest.raises(KeyError):
            alchemy_store.facts._update(fact)

    def test_update_fails_pk_none(self, alchemy_store, alchemy_fact, new_fact_values):
        """Make sure that trying to update a new fact ,e.g. one without a pk."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        new_values = new_fact_values(fact)
        fact.start = new_values["start"]
        fact.end = new_values["end"]
        fact.pk = None
        with pytest.raises(ValueError):
            alchemy_store.facts._update(fact)

    # ***

    def test_save_new(self, fact, alchemy_store):
        count_before = alchemy_store.session.query(AlchemyFact).count()
        result = alchemy_store.facts.save(fact)
        count_after = alchemy_store.session.query(AlchemyFact).count()
        assert count_before < count_after
        assert result.activity.name == fact.activity.name
        assert result.description == fact.description

    # ***

    def test_remove_normal(self, alchemy_store, alchemy_fact):
        """Make sure the fact but not its tags are removed."""
        count_before = alchemy_store.session.query(AlchemyFact).count()
        tags_before = alchemy_store.session.query(AlchemyTag).count()
        fact = alchemy_fact.as_hamster(alchemy_store)
        alchemy_store.facts.remove(fact, purge=True)
        count_after = alchemy_store.session.query(AlchemyFact).count()
        assert count_after < count_before
        assert alchemy_store.session.query(AlchemyFact).get(fact.pk) is None
        assert alchemy_store.session.query(AlchemyTag).count() == tags_before

    def test_remove_fails_no_pk(self, alchemy_store, alchemy_fact):
        """Ensure remove() raises ValueError when passed Fack without PK."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        fact.pk = None
        with pytest.raises(ValueError):
            alchemy_store.facts.remove(fact)

    def test_remove_fails_unknown_pk(self, alchemy_store, alchemy_fact):
        """Ensure remove() raises ValueError when passed Fack with unknown PK."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        fact.pk += 1
        with pytest.raises(KeyError):
            alchemy_store.facts.remove(fact)

    def test_remove_fails_already_deleted(self, alchemy_store, alchemy_fact):
        """Ensure remove() raises ValueError when passed Fack marked deleted."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        fact.deleted = True
        alchemy_store.facts.save(fact)
        with pytest.raises(Exception):
            alchemy_store.facts.remove(fact)

    # ***

    def test_get(self, alchemy_store, alchemy_fact):
        fact = alchemy_fact.as_hamster(alchemy_store)
        result = alchemy_store.facts.get(fact.pk)
        assert result == fact

    # Most of the get_all tests are in test_gather_fact, except this one.
    @freeze_time("2015-12-12 18:00")
    def test_get_all_since_until(
        self,
        alchemy_store,
        mocker,
        fact,
        search_parameter_parametrized,
    ):
        """Ensure since and until are converted to datetime for backend function."""
        # See also: nark's test_get_all_various_since_and_until_times
        since, until, description, expectation = search_parameter_parametrized
        mocker.patch.object(alchemy_store.facts, "gather", return_value=[fact])
        # F841 local variable '_facts' is assigned to but never used
        _facts = alchemy_store.facts.get_all(since=since, until=until)  # noqa: F841
        assert alchemy_store.facts.gather.called
        # call_args is (args, kwargs), and QueryTerms is the first args arg.
        query_terms = alchemy_store.facts.gather.call_args[0][0]
        assert query_terms.since == expectation["since"]
        assert query_terms.until == expectation["until"]

    # ***

    def test_starting_at(self, alchemy_store, alchemy_fact):
        """Verify FactManager.starting_at."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        # Ensure that starting_at does not exclude the reference Fact.
        ref = fact.copy(include_pk=False)
        result = alchemy_store.facts.starting_at(ref)
        assert result == fact

    # ***

    def test_ending_at(self, alchemy_store, alchemy_fact):
        """Verify FactManager.ending_at."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        # Ensure that starting_at does not exclude the reference Fact.
        ref = fact.copy(include_pk=False)
        result = alchemy_store.facts.ending_at(ref)
        assert result == fact

    # ***

    @pytest.mark.parametrize("send_fact", (False, "closed", "active"))
    def test_antecedent(self, alchemy_store, set_of_alchemy_facts_active, send_fact):
        """Verify FactManager.antecedent works with various reference input."""
        assert len(set_of_alchemy_facts_active) == 5
        # Depending on send_fact, either send the fact, or a datetime.
        if send_fact:
            ref_time = None
            if send_fact == "closed":
                fact = set_of_alchemy_facts_active[3]
                expect = set_of_alchemy_facts_active[2]
            else:  # 'active'
                fact = set_of_alchemy_facts_active[4]
                expect = set_of_alchemy_facts_active[3]
        else:
            fact = None
            ref_time = set_of_alchemy_facts_active[3].start
            expect = set_of_alchemy_facts_active[2]
        result = alchemy_store.facts.antecedent(fact=fact, ref_time=ref_time)
        assert result == expect

    # ***

    @pytest.mark.parametrize("send_fact", (False, True))
    def test_subsequent(self, alchemy_store, set_of_alchemy_facts, send_fact):
        """Verify FactManager.subsequent works with various reference input."""
        assert len(set_of_alchemy_facts) == 5
        fact_2nd = set_of_alchemy_facts[1]
        fact_3rd = set_of_alchemy_facts[2]
        # Depending on send_fact, either send the fact, or a datetime.
        fact = fact_2nd if send_fact else None
        ref_time = fact_2nd.end if not send_fact else None
        result = alchemy_store.facts.subsequent(fact=fact, ref_time=ref_time)
        # We could instead checked against a real Fact, either works:
        #  fact_3rd = set_of_alchemy_facts[2].as_hamster(alchemy_store)
        assert result == fact_3rd

    # ***

    def test_strictly_during(self, alchemy_store, set_of_alchemy_facts):
        """Verify FactManager.strictly_during finds a range of Facts."""
        assert len(set_of_alchemy_facts) == 5
        expect = set_of_alchemy_facts[1:2]
        results = alchemy_store.facts.strictly_during(
            since=expect[0].start,
            until=expect[-1].end,
        )
        assert results == expect

    # ***

    def test_surrounding_exclusive_outer(self, alchemy_store, set_of_alchemy_facts):
        """Verify exclusive surrounding finds nothing given a Fact's start or end."""
        assert len(set_of_alchemy_facts) == 5
        any_fact = set_of_alchemy_facts[2]
        fact_time = any_fact.start
        results = alchemy_store.facts.surrounding(fact_time=fact_time, inclusive=False)
        assert results == []

    def test_surrounding_exclusive_inner(self, alchemy_store, set_of_alchemy_facts):
        """Verify exclusive surrounding finds Fact given time between start and end."""
        assert len(set_of_alchemy_facts) == 5
        any_fact = set_of_alchemy_facts[2]
        fact_time = any_fact.end - ((any_fact.end - any_fact.start) / 2)
        results = alchemy_store.facts.surrounding(fact_time=fact_time, inclusive=False)
        assert results[0] == any_fact

    def test_surrounding_inclusive_outer(
        self,
        alchemy_store,
        set_of_alchemy_facts_contiguous,
    ):
        """Verify inclusive surrounding finds 2 Facts given time at end and start."""
        assert len(set_of_alchemy_facts_contiguous) == 5
        expect = [
            set_of_alchemy_facts_contiguous[1],
            set_of_alchemy_facts_contiguous[2],
        ]
        fact_time = expect[-1].start
        results = alchemy_store.facts.surrounding(fact_time=fact_time, inclusive=True)
        assert results == expect

    def test_surrounding_inclusive_inner(self, alchemy_store, set_of_alchemy_facts):
        """Verify inclusive surrounding finds Fact given time between start and end."""
        assert len(set_of_alchemy_facts) == 5
        any_fact = set_of_alchemy_facts[2]
        fact_time = any_fact.end - ((any_fact.end - any_fact.start) / 2)
        results = alchemy_store.facts.surrounding(fact_time=fact_time, inclusive=True)
        assert results[0] == any_fact

    # ***

    def test_endless(self, alchemy_store, set_of_alchemy_facts_active):
        """Verify FactManager.endless finds the active Fact."""
        assert len(set_of_alchemy_facts_active) == 5
        expect = set_of_alchemy_facts_active[-1]
        results = alchemy_store.facts.endless()
        assert results[0] == expect

    # ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest

from nark.backends.sqlalchemy.objects import AlchemyTag


class TestTagManager:
    """"""

    def test_add_new(self, alchemy_store, alchemy_tag_factory):
        """
        Our manager methods return the persistant instance as hamster objects.
        As we want to make sure that we compare our expectations against the
        raw saved object, we look it up explicitly again.
        """
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory.build().as_hamster(alchemy_store)
        tag.pk = None
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        result = alchemy_store.tags._add(tag)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        db_instance = alchemy_store.session.query(AlchemyTag).get(result.pk)
        assert tag.equal_fields(db_instance)
        assert tag != db_instance

    def test_add_existing_name(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that adding a tag with a name that is already present
        gives an error.
        """
        existing_tag = alchemy_tag_factory()
        tag = alchemy_tag_factory.build().as_hamster(alchemy_store)
        tag.name = existing_tag.name
        tag.pk = None
        with pytest.raises(ValueError):
            alchemy_store.tags._add(tag)

    def test_add_with_pk(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that adding a alchemy_tag that already got an PK raises
        an exception.
        """
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        tag.name += "foobar"
        assert tag.pk
        with pytest.raises(ValueError):
            alchemy_store.tags._add(tag)

    def test_update(self, alchemy_store, alchemy_tag_factory, new_tag_values):
        """Test that updateing a alchemy_tag works as expected."""
        alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        new_values = new_tag_values(tag)
        for key, value in new_values.items():
            assert getattr(tag, key) != value
        for key, value in new_values.items():
            setattr(tag, key, value)
        alchemy_store.tags._update(tag)
        db_instance = alchemy_store.session.query(AlchemyTag).get(tag.pk)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        assert tag.equal_fields(db_instance)

    def test_update_without_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure that passing a tag without a PK raises an error."""
        tag = alchemy_tag_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.tags._update(tag)

    def test_update_invalid_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure that passing a tag with a non existing PK raises an error."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        tag.pk = tag.pk + 10
        with pytest.raises(KeyError):
            alchemy_store.tags._update(tag)

    def test_update_existing_name(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that renaming a given alchemy_tag to a taken name throws an error.
        """
        tag_1, tag_2 = (alchemy_tag_factory(), alchemy_tag_factory())
        tag_2 = tag_2.as_hamster(alchemy_store)
        tag_2.name = tag_1.name
        with pytest.raises(ValueError):
            alchemy_store.tags._update(tag_2)

    def test_remove(self, alchemy_store, alchemy_tag_factory):
        """Make sure passing a valid alchemy_tag removes it from the db."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        alchemy_store.tags.remove(tag)
        assert alchemy_store.session.query(AlchemyTag).get(tag.pk) is None

    def test_remove_no_pk(self, alchemy_store, alchemy_tag_factory):
        """Ensure that passing a alchemy_tag without an PK raises an error."""
        tag = alchemy_tag_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.tags.remove(tag)

    def test_remove_invalid_pk(self, alchemy_store, alchemy_tag_factory):
        """Ensure that passing a alchemy_tag without an PK raises an error."""
        tag = alchemy_tag_factory.build(pk=800).as_hamster(alchemy_store)
        with pytest.raises(KeyError):
            alchemy_store.tags.remove(tag)

    def test_get_existing_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure method retrieves corresponding object."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.get(tag.pk)
        assert result == tag

    def test_get_non_existing_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure we throw an error if PK can not be resolved."""
        alchemy_store.session.query(AlchemyTag).count == 0
        tag = alchemy_tag_factory()
        alchemy_store.session.query(AlchemyTag).count == 1
        with pytest.raises(KeyError):
            alchemy_store.tags.get(tag.pk + 1)

    def test_get_by_name(self, alchemy_tag_factory, alchemy_store):
        """Make sure a alchemy_tag can be retrieved by name."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.get_by_name(tag.name)
        assert result == tag

    def test_get_all(self, alchemy_store, set_of_tags):
        results = alchemy_store.tags.get_all()
        assert len(results) == len(set_of_tags)
        assert len(results) == alchemy_store.session.query(AlchemyTag).count()
        for tag in set_of_tags:
            assert tag.as_hamster(alchemy_store) in results

    # Test convenience methods.
    def test_get_or_create_get(self, alchemy_store, alchemy_tag_factory):
        """
        Test that if we pass a alchemy_tag of existing name, we just return it.
        """
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.get_or_create(tag)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        assert result == tag

    def test_get_or_create_new_name(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that passing a tag with new name creates and returns new instance.
        """
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory.build().as_hamster(alchemy_store)
        tag.pk = None
        result = alchemy_store.tags.get_or_create(tag)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        assert result.equal_fields(tag)

    def test_get_deleted_item(self, alchemy_store, alchemy_tag):
        """Make sure method retrieves deleted object."""
        alchemy_tag.deleted = True
        result = alchemy_store.tags.get(alchemy_tag.pk, deleted=True)
        assert result == alchemy_tag

    def test_get_all_match_tags(self, alchemy_store, set_of_alchemy_facts_active):
        """Test get_all argument: QueryTerms.match_tags."""
        # Note that non-Fact query does not actually implement the
        # QueryTerms.query_filter_by_tags option (because no obvious
        # use case). But QueryTerms still accepts it.
        results = alchemy_store.tags.get_all(
            match_tags=["foo"],
        )
        expect_all_tags = 0
        for fact in set_of_alchemy_facts_active:
            expect_all_tags += len(fact.tags)
        assert len(results) == expect_all_tags

    # ***

    @pytest.mark.parametrize(
        ("sort_cols"),
        (
            (["start"]),
            ([None]),
            (["usage"]),
            (["time"]),
            (["activity"]),
            (["category"]),
            (["tag"]),
        ),
    )
    def test_get_all_sort_cols(self, alchemy_store, sort_cols):
        alchemy_store.tags.get_all(sort_cols=sort_cols)

    def test_get_all_sort_cols_unknown(self, alchemy_store, mocker):
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.tags.get_all(sort_cols=["foo"])
        assert alchemy_store.logger.warning.called



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest

from nark.backends.sqlalchemy.managers.fact import FactManager
from nark.backends.sqlalchemy.managers.gather_fact import GatherFactManager
from nark.backends.sqlalchemy.objects import AlchemyFact
from nark.items.fact import Fact
from nark.items.tag import Tag


class TestGatherFactManager:
    """"""

    def test_get_all_argless(self, set_of_alchemy_facts, alchemy_store):
        results = alchemy_store.facts.get_all()
        assert len(results) == len(set_of_alchemy_facts)
        assert len(results) == alchemy_store.session.query(AlchemyFact).count()
        assert isinstance(results[0], Fact)

    def test_get_all_raw_lazy_tags(self, set_of_alchemy_facts, alchemy_store):
        results = alchemy_store.facts.get_all(raw=True, lazy_tags=True)
        assert len(results) == len(set_of_alchemy_facts)
        # (lb): This raises IntegrityError... on a call to `INSERT INTO tags`...
        # not really sure what's up, don't care too much, but this query works
        # in the test_get_all_argless test, where raw=False, but with raw=True,
        # it fails. Something wrong with our db session?
        #   assert len(results) == alchemy_store.session.query(AlchemyFact).count()
        assert isinstance(results[0], AlchemyFact)

    @pytest.mark.parametrize(
        ("start_filter", "end_filter"),
        (
            (10, 12),
            (10, None),
            (None, -12),
        ),
    )
    def test_get_all_existing_facts_not_in_timerange(
        self,
        alchemy_store,
        alchemy_fact,
        bool_value_parametrized,
        start_filter,
        end_filter,
    ):
        """Make sure that a valid timeframe returns an empty list."""
        since, until = None, None
        if start_filter:
            since = alchemy_fact.start + datetime.timedelta(days=start_filter)
        if end_filter:
            until = alchemy_fact.start + datetime.timedelta(days=end_filter)

        results = alchemy_store.facts.get_all(
            since=since,
            until=until,
            partial=bool_value_parametrized,
        )
        assert results == []

    @pytest.mark.parametrize(
        ("start_filter", "end_filter"),
        (
            (-1, 5),
            (-1, None),
            (None, 5),
            (None, None),
        ),
    )
    def test_get_all_existing_fact_fully_in_timerange(
        self,
        alchemy_store,
        alchemy_fact,
        bool_value_parametrized,
        start_filter,
        end_filter,
    ):
        """Ensure a fact fully within the timeframe is returned."""
        since, until = None, None
        if start_filter:
            since = alchemy_fact.start + datetime.timedelta(days=start_filter)
        if end_filter:
            until = alchemy_fact.start + datetime.timedelta(days=end_filter)

        results = alchemy_store.facts.get_all(
            since=since,
            until=until,
            partial=bool_value_parametrized,
        )
        # ANSWER/2018-05-05: (lb): This test is failing. When did it break?
        #                      assert results == [alchemy_fact]
        #                    - 2020-05-26: Probably when lazy_tags=False added.
        assert len(results) == 1
        assert str(results[0]) == str(alchemy_fact)

    @pytest.mark.parametrize(
        ("start_filter", "end_filter"),
        (
            # Fact.start is in time window.
            (None, 2),
            (-900, 2),
            # Fact.end is in time window.
            (5, None),
            (5, 900),
        ),
    )
    def test_get_all_existing_fact_partialy_in_timerange(
        self,
        alchemy_store,
        alchemy_fact,
        bool_value_parametrized,
        start_filter,
        end_filter,
    ):
        """
        Test that a fact partially within timeframe is returned with
        ``partial=True`` only.
        """
        since, until = None, None
        if start_filter:
            since = alchemy_fact.start + datetime.timedelta(minutes=start_filter)
        if end_filter:
            until = alchemy_fact.start + datetime.timedelta(minutes=end_filter)
        results = alchemy_store.facts.get_all(
            since=since,
            until=until,
            partial=bool_value_parametrized,
            # Use lazy_tags=True, so that we can compare Fact objects.
            # - If lazy_tags=False, the query concatenates tag names to save
            #   time, and does not load Tag PKs, and each Fact object is setup
            #   with new Tags that do not have their PK set. As such, the
            #   comparison, `results == [alchemy_fact]`, won't work.
            #   (Alternatively, one could compare str(fact), which does not
            #   include IDs.)
            # - But when lazy_tags=True, this relies on SQLAlchemy magic that
            #   fetches a Fact's Tag items when the fact.tags attribute is read.
            #   It's slower (there's one additional SELECT for every fact.tags
            #   read) but more complete.
            lazy_tags=True,
        )
        if bool_value_parametrized:
            assert len(results) == 1
            assert str(results[0]) == str(alchemy_fact)
            assert results == [alchemy_fact]
        else:
            assert results == []

    def test_get_all_search_matches_description(
        self,
        alchemy_store,
        set_of_alchemy_facts,
    ):
        """Make sure facts with ``Fact.activity.name`` matching term are returned."""
        assert len(set_of_alchemy_facts) == 5
        search_terms = [set_of_alchemy_facts[1].description]
        # Use lazy_tags=True so Tag.pk are set, and results == [...] works.
        results = alchemy_store.facts.get_all(
            search_terms=search_terms,
            lazy_tags=True,
        )
        assert len(results) == 1
        assert str(results[0]) == str(set_of_alchemy_facts[1])
        assert results == [set_of_alchemy_facts[1]]

    def test_get_all_search_matches_activity(self, alchemy_store, set_of_alchemy_facts):
        """Make sure facts with ``Fact.activity.name`` matching term are returned."""
        assert len(set_of_alchemy_facts) == 5
        search_terms = [set_of_alchemy_facts[1].activity.name]
        # Use lazy_tags=True so Tag.pk are set, and results == [...] works.
        results = alchemy_store.facts.get_all(
            search_terms=search_terms,
            broad_match=True,
            lazy_tags=True,
        )
        assert len(results) == 1
        assert str(results[0]) == str(set_of_alchemy_facts[1])
        assert results == [set_of_alchemy_facts[1]]

    def test_get_all_search_matches_category(self, alchemy_store, set_of_alchemy_facts):
        """Make sure facts with ``Fact.category.name`` matching term are returned."""
        assert len(set_of_alchemy_facts) == 5
        search_terms = [set_of_alchemy_facts[1].category.name]
        # Use lazy_tags=True so Tag.pk are set, and results == [...] works.
        results = alchemy_store.facts.get_all(
            search_terms=search_terms,
            broad_match=True,
            lazy_tags=True,
        )
        assert len(results) == 1
        assert str(results[0]) == str(set_of_alchemy_facts[1])
        assert results == [set_of_alchemy_facts[1]]

    # ***

    def test__get_all_no_query_terms_not_lazy(
        self, alchemy_store, set_of_alchemy_facts
    ):
        """Verify basic FactManager.get_all finds the whole store."""
        assert len(set_of_alchemy_facts) == 5
        results = alchemy_store.facts.get_all(lazy_tags=False)
        assert len(results) == len(set_of_alchemy_facts)
        # (lb): The results list contain the same items, and the sort order is
        # the same, because the Fact.get_all sort defaults to 'start'. But the
        # lists are not equal, because the result Facts do not have the Tag PK.
        # That is, lazy_tags enables a quicker query that just concatenates tag
        # names, not bothering with the PK. So comparing lists will not work:
        #   assert results == set_of_alchemy_facts
        # but we can stringify and compare.
        assert str(results[0]) == str(set_of_alchemy_facts[0])
        # etc.

    def test__get_all_no_query_terms_yes_lazy(
        self,
        alchemy_store,
        set_of_alchemy_facts,
    ):
        """Verify basic FactManager.get_all finds the whole store."""
        assert len(set_of_alchemy_facts) == 5
        # Use lazy_tags=True so Tag.pk are set, and results == ... works.
        results = alchemy_store.facts.get_all(lazy_tags=True)
        # (lb): This is similar to the previous test, but now lazy_tags is True,
        # meaning, whenever we access a Tag item for the first time, SQLAlchemy
        # will fetch it from the database. But it also means that the Fact.tags
        # include PKs, so here, a direct list comparison is valid.
        assert len(results) == len(set_of_alchemy_facts)
        assert results == set_of_alchemy_facts

    def test__get_all_no_params_count(self, alchemy_store, set_of_alchemy_facts):
        """Verify FactManager.get_all count_results returns number of Facts."""
        assert len(set_of_alchemy_facts) == 5
        results = alchemy_store.facts.get_all(count_results=True)
        assert results == 5

    def test__get_all_include_stats_return_raw(
        self,
        alchemy_store,
        set_of_alchemy_facts,
    ):
        """Verify basic FactManager.get_all finds the whole store."""
        assert len(set_of_alchemy_facts) == 5
        results = alchemy_store.facts.get_all(
            include_stats=True,
            raw=True,
        )
        assert len(results) == 5
        # Each results is the Fact and the aggregate columns.
        fact_0, *cols_0 = results[0]
        assert len(cols_0) == len(FactManager.RESULT_GRP_INDEX)
        # Duration is days, and set_of_alchemy_facts are 20 mins. each.
        i_duration = FactManager.RESULT_GRP_INDEX["duration"]
        # Ensure that the duration calculation is correct. (For instance,
        # if the SUM() aggregate happened before tags are coalesced, if a
        # Fact had three tags, the duration calculation would be three times
        # would it really should be.)
        assert round(cols_0[i_duration] * 24 * 60) == 20
        # The group_count when not grouping is 1. (And even if we did group,
        # faker seems to make it so all Activity, Category, and Tag names
        # are unique.)
        # FIXME/2020-05-25: Create Facts with the same Activity, and test aggregates.
        i_group_count = FactManager.RESULT_GRP_INDEX["group_count"]
        assert cols_0[i_group_count] == 1
        # Because raw=True, the Tag objects were recreated, but without PKs.
        assert not fact_0.tags[0].pk

    def test__get_all_by_pk_with_stats_named(self, alchemy_store, set_of_alchemy_facts):
        """Verify QueryTerms.named_tuples returns attribute-accessible results."""
        assert len(set_of_alchemy_facts) == 5
        expect = set_of_alchemy_facts[3]
        results = alchemy_store.facts.get_all(
            named_tuples=True,
            include_stats=True,
            key=expect.pk,
        )
        assert len(results) == 1
        assert str(results[0].fact) == str(expect)
        assert results[0].group_count == 1
        # etc.

    def test__get_all_limit_offset(self, alchemy_store, set_of_alchemy_facts):
        """Verify FactManager.get_all count_results returns number of Facts."""
        assert len(set_of_alchemy_facts) == 5
        results = alchemy_store.facts.get_all(limit=2, offset=2, lazy_tags=True)
        assert len(results) == 2
        assert results == set_of_alchemy_facts[2:4]

    # ***

    def test__get_all_fails_on_unsupported_store(self, controller, alchemy_store):
        """Ensure get_all/GatherFactManager.gather raised on unsupported store type."""
        # MAYBE/2020-05-26: Add support for other data stores (other than SQLite),
        #   which requires using and wiring the appropriate DBMS-specific aggregate
        #   functions.
        #   - For now, dob is at least nice enough to print an error message.
        controller.store.config["db.engine"] += "_not"
        with pytest.raises(NotImplementedError) as excinfo:
            results = alchemy_store.facts.get_all()  # noqa: F841 local variable...
            assert False  # Unreachable.
        # See: must_support_db_engine_funcs.
        expect = "This feature does not work with the current DBMS engine"
        assert str(excinfo.value).startswith(expect)

    # ***

    def test_get_all_match_tags(
        self,
        alchemy_store,
        set_of_alchemy_facts_active,
        alchemy_tag_factory,
    ):
        """Test get_all argument: QueryTerms.match_activities."""
        tag_0_0 = set_of_alchemy_facts_active[0].tags[0]
        tag_2_2 = set_of_alchemy_facts_active[2].tags[2].name
        # This Tag will not be found because not associated with any Fact.
        any_tag = alchemy_tag_factory(pk=None)
        Tag_with_pk_None = Tag(name=any_tag.name)
        results = alchemy_store.facts.get_all(
            match_tags=[
                tag_0_0,
                tag_2_2,
                Tag_with_pk_None,
                None,
            ],
        )
        # Given tag_0_0 and tag_2_2, expect 2 matching Facts.
        assert len(results) == 2

    # ***

    def test_get_all_include_stats_facts(self, alchemy_store, set_of_alchemy_facts):
        results = alchemy_store.facts.get_all(include_stats=True)
        # We didn't aggregate results, so they're all returned.
        assert len(results) == len(set_of_alchemy_facts)
        fact_stats = GatherFactManager.FactStatsTuple(*results[0])
        assert fact_stats.group_count == 1

    # ***

    @pytest.mark.parametrize(
        ("group_activity", "group_category", "group_tags", "group_days"),
        (
            (True, False, False, False),
            (False, True, False, False),
            (True, True, False, False),
            (False, False, True, False),
            (False, False, False, True),
        ),
    )
    def test_get_all_prepare_actg_cols(
        self,
        alchemy_store,
        # The first fixture we made for tests makes 5 Facts each 1 day apart:
        #   set_of_alchemy_facts
        # but if we get facts on the same day, we can test group_days works.
        set_of_alchemy_facts_contiguous,
        group_activity,
        group_category,
        group_tags,
        group_days,
    ):
        results = alchemy_store.facts.get_all(
            group_activity=group_activity,
            group_category=group_category,
            group_tags=group_tags,
            group_days=group_days,
            # Included stats so _process_record_reduce_aggregate_value
            # called with group_concat values (and we get coverage of
            # encoded_value.split(magic_tag_sep)).
            include_stats=True,
        )
        # (lb): Not sure I care enough to really test that grouping works,
        # i.e., set Facts' activities the same, etc.
        if not group_days:
            expect_count = len(set_of_alchemy_facts_contiguous)
        else:
            expect_count = 1
        assert len(results) == expect_count

    def test_get_all_group_by_tags_lazy_tags_raises(self, alchemy_store):
        with pytest.raises(Exception):
            alchemy_store.facts.get_all(group_tags=True, lazy_tags=True)

    def test_get_all_group_by_actegories_nameless(
        self,
        alchemy_store,
        alchemy_fact_factory,
    ):
        alchemy_fact_1 = alchemy_fact_factory()
        alchemy_fact_2 = alchemy_fact_factory()
        alchemy_fact_1.activity.name = ""
        alchemy_fact_2.activity = alchemy_fact_1.activity
        results = alchemy_store.facts.get_all(
            group_category=True,
            include_stats=True,
        )
        assert len(results) == 1

    def test_get_all_group_by_activities_nameless(
        self,
        alchemy_store,
        alchemy_fact_factory,
    ):
        # (lb): Just for one line of coverage in _process_record_reduce_aggregate_value:
        #           else:
        #               unique_values = ''
        alchemy_fact_1 = alchemy_fact_factory()
        alchemy_fact_2 = alchemy_fact_factory()
        # Clear activity names so the `if encoded_value` is falsey (empty string).
        alchemy_fact_1.activity.name = ""
        alchemy_fact_2.activity.name = ""
        results = alchemy_store.facts.get_all(
            # Note that we group by category, not activity, so that the activity
            # names are group_concat'ed.
            #  NOPE: group_activity=True,
            group_category=True,
            # Ask for stats to have function called that has line we want to test.
            include_stats=True,
        )
        # We left category names unique, so will get two categories.
        # - But at least that line we want to test gets tested.
        assert len(results) == 2

    # ***

    def test_get_all_exclude_ongoing(
        self,
        alchemy_store,
        set_of_alchemy_facts_active,
    ):
        results = alchemy_store.facts.get_all(exclude_ongoing=True)
        assert len(results) == len(set_of_alchemy_facts_active) - 1

    # ***

    @pytest.mark.parametrize(
        ("sort_cols"),
        (
            (["start"]),
            ([None]),
            (["time"]),
            (["day"]),
            (["activity"]),
            (["category"]),
            (["tag"]),
            (["usage"]),
            (["name"]),
            (["fact"]),
        ),
    )
    def test_get_all_sort_cols(self, alchemy_store, sort_cols):
        alchemy_store.facts.get_all(sort_cols=sort_cols)

    def test_get_all_sort_cols_unknown(self, alchemy_store, mocker):
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.facts.get_all(sort_cols=["foo"])
        assert alchemy_store.logger.warning.called



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Test for report writer module."""

import os.path
import sys

import pytest

from nark.reports import ReportWriter


class TestReportWriter(object):
    @pytest.mark.parametrize(
        ["datetime_format", "expected_format"],
        [
            [None, "%Y-%m-%d %H:%M:%S"],
            ["%Y-%m-%d", "%Y-%m-%d"],
        ],
    )
    def test_report_writer_init_stores_datetime_format(
        self,
        path,
        datetime_format,
        expected_format,
    ):
        """Make sure that Writer initialization stores the ``datetime_format``."""
        report_writer = ReportWriter()
        report_writer.output_setup(path, datetime_format=datetime_format)
        assert report_writer.datetime_format == expected_format

    def test_report_writer_output_setup_opens_file_path_given(self, path):
        """Make sure a file like object is beeing opened."""
        report_writer = ReportWriter()
        report_writer.output_setup(path)
        assert os.path.isfile(path)
        assert report_writer.output_file.closed is False

    def test_report_writer_output_setup_falls_back_to_stdout(self):
        """Make sure stdout used if no file path specified."""
        report_writer = ReportWriter()
        report_writer.output_setup(output_obj="")
        assert report_writer.output_file is sys.stdout

    def test_report_writer_output_setup_opens_binarily(self, path):
        """Make sure a file is opened for binary when specified."""
        # 2020-06-08: output_b was only used for Python 2 xml, but py2 support
        # long since dropped, so this is the only code that uses output_b.
        report_writer = ReportWriter(output_b=True)
        report_writer.output_setup(path)
        assert report_writer.output_file.mode == "wb"

    def test_report_writer_output_setup_uses_object_pass(self):
        """Make sure stdout used if no file path specified."""
        report_writer = ReportWriter()
        not_a_string_path = object()
        report_writer.output_setup(output_obj=not_a_string_path)
        assert report_writer.output_file is not_a_string_path

    def test_report_writer_write_facts_calls__write_fact(
        self,
        mocker,
        report_writer,
        list_of_facts,
    ):
        """Make sure that each ``Fact`` instances triggers a new line."""
        number_of_facts = 10
        facts = list_of_facts(number_of_facts)
        mocker.patch.object(report_writer, "_write_fact", return_value=None)
        report_writer.write_facts(facts)
        assert report_writer._write_fact.call_count == number_of_facts
        # For coverage!
        assert report_writer.requires_table is False

    def test_report_writer_write_facts_respects_row_limit(
        self,
        mocker,
        report_writer,
        list_of_facts,
    ):
        """Ensure that write_facts respects the row_limit."""
        row_limit = 3
        number_of_facts = 10
        facts = list_of_facts(number_of_facts)
        report_writer.row_limit = row_limit
        mocker.patch.object(report_writer, "_write_fact", return_value=None)
        report_writer.write_facts(facts)
        assert report_writer._write_fact.call_count == row_limit

    def test_report_writer_write_facts_fails_to__close(
        self,
        report_writer,
        list_of_facts,
    ):
        """Make sure our output file is closed at the end."""
        number_of_facts = 10
        facts = list_of_facts(number_of_facts)
        assert report_writer.output_file.closed is False
        # Because we do not mock _write_fact, it raises.
        with pytest.raises(NotImplementedError):
            report_writer.write_facts(facts)
        assert report_writer.output_file.closed is False

    def test_report_writer_write_report_calls__write_result(
        self,
        mocker,
        report_writer,
        table,
        headers,
    ):
        mocker.patch.object(report_writer, "_write_result", return_value=None)
        report_writer.write_report(table, headers)
        assert report_writer._write_result.call_count == len(table)

    def test_report_writer_write_report_respects_row_limit(
        self,
        mocker,
        report_writer,
        table,
        headers,
    ):
        """Ensure that write_report respects the row_limit."""
        row_limit = 3
        report_writer.row_limit = row_limit
        mocker.patch.object(report_writer, "_write_result", return_value=None)
        report_writer.write_report(table, headers)
        assert report_writer._write_result.call_count == row_limit

    def test_report_writer_write_report_fails_to__close(
        self,
        report_writer,
        table,
        headers,
    ):
        assert report_writer.output_file.closed is False
        # Because we do not mock _write_result, it raises.
        with pytest.raises(NotImplementedError):
            report_writer.write_report(table, headers)
        assert report_writer.output_file.closed is False

    def test_report_writer__close(self, report_writer, path):
        """Ensure that the the output gets closed."""
        report_writer._close()
        assert report_writer.output_file.closed
